"""Subcommand for importing translations."""

import os
import click
import glob
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.translations.import')


@click.command(name='import')
@shared.pod_path_argument
@click.option('--source', type=click.Path(), required=True,
              help='Path to source (either zip file, directory, or file).')
@shared.include_obsolete_option(CFG)
@shared.locale_option(
    help_text='Locale of the message catalog to import. This option is'
              ' only applicable when --source is a .po file.', multiple=False)
@click.option('--untranslated', default=False, is_flag=True,
              help='Whether to only import untranslated strings.')
def translations_import(pod_path, source, locale, include_obsolete, untranslated):
    """Imports translations from an external source."""
    if source.endswith('.po') and locale is None:
        text = 'Must specify --locale when --source is a .po file.'
        raise click.ClickException(text)
    if not source.endswith('.po') and locale is not None:
        text = 'Cannot specify --locale when --source is not a .po file.'
        raise click.ClickException(text)
    source = os.path.expanduser(source)
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    if not pod.exists:
        raise click.ClickException('Pod does not exist: {}'.format(pod.root))
    source = glob.glob(source)
    with pod.profile.timer('translations_grow_i'):
        for path in source:
            pod.catalogs.import_translations(
                path, locale=locale, include_obsolete=include_obsolete,
                untranslated=untranslated)
    return pod



from protorpc import messages


class Result(messages.Enum):
    PASS = 1
    FAIL = 2
    WARNING = 3


class TestResultMessage(messages.Message):
    result = messages.EnumField(Result, 1, default=Result.PASS)
    title = messages.StringField(2)
    text = messages.StringField(3)


class TestResultsMessage(messages.Message):
    test_results = messages.MessageField(TestResultMessage, 1, repeated=True)



"""Common grow utility functions."""

import csv as csv_lib
import codecs
import fnmatch
import functools
import gettext
import json
import logging
import os
import re
import string
import sys
import threading
import time
from urllib import parse as url_parse
from collections import OrderedDict
import yaml
import bs4
import git
import html2text
import translitcodec  # pylint: disable=unused-import
from grow.common import structures
from grow.common import untag
from grow.common import yaml_utils
from grow.pods import errors

# The C (LibYAML) version is massively faster than the pure Python version.
# Since YAML performance is critical, display a warning message with information
# related to fixing it.
try:
    # pylint: disable=ungrouped-imports
    from yaml import CLoader as yaml_Loader
except ImportError:
    logging.warning('Warning: libyaml missing, using slower yaml parser. See https://grow.dev/libyaml')
    from yaml import Loader as yaml_Loader


LOCALIZED_KEY_REGEX = re.compile('(.*)@([^@]+)$')
SENTINEL = object()
DRAFT_KEY = '$draft'
SLUG_REGEX = re.compile(r'[\t !"#$%&\'()*\-/<=>?@\[\\\]^_`{|},\.;]+')
SLUG_SUBSTITUTE = ((':{}', ':'),)


class Error(Exception):
    """Base error class."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class UnavailableError(Error):
    """Raised when a feature is not available."""
    pass


class DraftStringError(Error):
    """Raised when a draft string is used yet not allowed."""
    pass


def get_grow_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))


def get_git_repo(root):
    try:
        return git.Repo(root)
    except git.InvalidGitRepositoryError:
        logging.info('WARNING: No Git repository found in {}'.format(root))


def interactive_confirm(message, default=False, input_func=None):
    if input_func is None:
        input_func = lambda m: input(m).lower()

    choices = 'Y/n' if default is True else 'y/N'
    message = '{} [{}]: '.format(message, choices)
    choice = input_func(message)
    if choice == 'y':
        return True
    elif choice == 'n':
        return False
    return default


FORMATTER = string.Formatter()
def safe_format(base_string, *args, **kwargs):
    """Safely format a string using the modern string formatting with fallback."""
    safe_kwargs = structures.SafeDict(**kwargs)
    try:
        return FORMATTER.vformat(base_string, args, safe_kwargs)
    except TypeError as e:
        return FORMATTER.vformat(str(base_string), args, safe_kwargs)


def walk(node, callback, parent_key=None, parent_node=None):
    if node is None:
        return
    for key in node:
        if isinstance(node, dict):
            item = node[key]
        else:
            item = key

        if isinstance(node, (dict)) and isinstance(item, (list, set)):
            parent_key = key

        if isinstance(item, (list, set, dict)):
            walk(item, callback, parent_key=parent_key, parent_node=node)
        else:
            if isinstance(node, (list, set)):
                key = parent_key
            callback(item, key, node, parent_node=parent_node)


def validate_name(name):
    # TODO: better validation.
    if ('//' in name
        or '..' in name
            or ' ' in name):
        raise errors.BadNameError(
            'Names must be lowercase and only contain letters, numbers, '
            'backslashes, and dashes. Found: "{}"'.format(name))


class memoize(object):

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        key = (args, frozenset(list(kwargs.items())))
        try:
            return self.cache[key]
        except KeyError:
            value = self.func(*args, **kwargs)
            self.cache[key] = value
            return value
        except TypeError:
            return self.func(*args, **kwargs)

    def __repr__(self):
        return self.func.__doc__

    def __get__(self, obj, objtype):
        fn = functools.partial(self.__call__, obj)
        fn.reset = self._reset
        return fn

    def _reset(self):
        self.cache = {}


class cached_property(property):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):
            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work. Ported from `werkzeug`.
    """

    _sentinel = object()

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, self._sentinel)
        if value is self._sentinel:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


class memoize_tag(memoize):

    def __call__(self, *args, **kwargs):
        use_cache = kwargs.pop('use_cache', False)
        if use_cache is True:
            return super(memoize_tag, self).__call__(*args, **kwargs)
        return self.func(*args, **kwargs)


def every_two(l):
    return list(zip(l[::2], l[1::2]))


def make_base_yaml_loader(pod, locale=None, untag_params=None,
                          tracking_func=None):
    """Make a base yaml loader that does not touch collections or docs."""
    if not tracking_func:
        tracking_func = lambda *args, **kwargs: None

    # A default set of params for nested yaml parsing.
    if not untag_params and pod:
        untag_params = {
            'env': untag.UntagParamRegex(pod.env.name),
        }

    class BaseYamlLoader(yaml_Loader):

        @staticmethod
        def loader_locale():
            return locale

        @staticmethod
        def pod_path():
            return None

        @staticmethod
        def deep_reference(data, reference):
            data = structures.DeepReferenceDict(data)
            try:
                return data[reference]
            except KeyError:
                return None

        @staticmethod
        def read_csv(pod_path, locale):
            """Reads a csv file using a cache."""
            file_cache = pod.podcache.file_cache
            contents = file_cache.get(pod_path, locale=locale)
            if contents is None:
                contents = pod.read_csv(pod_path, locale=locale)
                contents = untag.Untag.untag(
                    contents, locale_identifier=locale, params=untag_params)
                file_cache.add(pod_path, contents, locale=locale)
            return contents

        @staticmethod
        def read_file(pod_path):
            """Reads a file using a cache."""
            file_cache = pod.podcache.file_cache
            contents = file_cache.get(pod_path)
            if contents is None:
                contents = pod.read_file(pod_path)
                file_cache.add(pod_path, contents)
            return contents

        @staticmethod
        def read_json(pod_path, locale):
            """Reads a json file using a cache."""
            file_cache = pod.podcache.file_cache
            contents = file_cache.get(pod_path, locale=locale)
            if contents is None:
                contents = pod.read_json(pod_path)
                contents = untag.Untag.untag(
                    contents, locale_identifier=locale, params=untag_params)
                file_cache.add(pod_path, contents, locale=locale)
            return contents

        @classmethod
        def read_string(cls, path):
            if '.' not in path:
                return None
            main, reference = path.split('.', 1)
            path = '/content/strings/{}.yaml'.format(main)
            tracking_func(path)
            if reference:
                data = structures.DeepReferenceDict(
                    cls.read_yaml(path, locale=cls.loader_locale()))
                try:
                    allow_draft = pod.podspec.fields.get('strings', {}).get('allow_draft')
                    if allow_draft is False and data.get(DRAFT_KEY):
                        raise DraftStringError('Encountered string in draft -> {}?{}'.format(path, reference))
                    value = data[reference]
                    if value is None:
                        if cls.pod_path():
                            pod.logger.warning(
                                'Missing {}.{} in {}'.format(
                                    main, reference, cls.pod_path()))
                        else:
                            pod.logger.warning(
                                'Missing {}.{}'.format(main, reference))
                    return value
                except KeyError:
                    return None
            return None

        @classmethod
        def read_yaml(cls, pod_path, locale):
            """Reads a yaml file using a cache."""
            file_cache = pod.podcache.file_cache
            contents = file_cache.get(pod_path, locale=locale)
            if contents is None:
                # Cannot use the file cache to store the raw data with the
                # `yaml.load` since constructors in the yaml loading are already
                # completed with the provided locale so untagged data is lost
                # and cannot be stored as raw data.
                contents = yaml.load(pod.read_file(pod_path), Loader=cls) or {}
                contents = untag.Untag.untag(
                    contents, locale_identifier=locale, params=untag_params)
                file_cache.add(pod_path, contents, locale=locale)
            return contents

        def _construct_func(self, node, func):
            if isinstance(node, yaml.SequenceNode):
                items = []
                for i, each in enumerate(node.value):
                    items.append(func(node.value[i].value))
                return items
            return func(node.value)

        def _track_dep_func(self, func):
            """Wrap a function with a call to the tracking function."""
            def _func(path, *args, **kwargs):
                tracking_func(path)
                return func(path, *args, **kwargs)
            return _func

        def construct_csv(self, node):
            def func(path):
                if '?' in path:
                    path, reference = path.split('?')
                    tracking_func(path)
                    return self.deep_reference(
                        self.read_csv(path, locale=self.loader_locale()), reference)
                tracking_func(path)
                return self.read_csv(path, locale=self.loader_locale())
            return self._construct_func(node, func)

        def construct_file(self, node):
            return self._construct_func(
                node, self._track_dep_func(self.read_file))

        def construct_gettext(self, node):
            return self._construct_func(node, gettext.gettext)

        def construct_json(self, node):
            def func(path):
                if '?' in path:
                    path, reference = path.split('?')
                    tracking_func(path)
                    return self.deep_reference(
                        self.read_json(path, locale=self.loader_locale()), reference)
                tracking_func(path)
                return self.read_json(path, locale=self.loader_locale())
            return self._construct_func(node, func)

        def construct_string(self, node):
            return self._construct_func(node, self.read_string)

        def construct_yaml(self, node):
            def func(path):
                if '?' in path:
                    path, reference = path.split('?')
                    tracking_func(path)
                    data = structures.DeepReferenceDict(
                        self.read_yaml(path, locale=self.loader_locale()))
                    try:
                        return data[reference]
                    except KeyError:
                        return None
                tracking_func(path)
                return self.read_yaml(path, locale=self.loader_locale())
            return self._construct_func(node, func)

    BaseYamlLoader.add_constructor('!_', BaseYamlLoader.construct_gettext)
    BaseYamlLoader.add_constructor('!g.csv', BaseYamlLoader.construct_csv)
    BaseYamlLoader.add_constructor('!g.file', BaseYamlLoader.construct_file)
    BaseYamlLoader.add_constructor('!g.json', BaseYamlLoader.construct_json)
    BaseYamlLoader.add_constructor('!g.string', BaseYamlLoader.construct_string)
    BaseYamlLoader.add_constructor('!g.yaml', BaseYamlLoader.construct_yaml)

    return BaseYamlLoader


def make_yaml_loader(pod, doc=None, locale=None, untag_params=None):
    # A default set of params for nested yaml parsing.
    if not untag_params and pod:
        untag_params = {
            'env': untag.UntagParamRegex(pod.env.name),
        }

    # Tracing function for dependency graph.
    tracking_func = lambda *args, **kwargs: None
    if pod and doc:
        # Add the path to the dependency graph in case it has no external refs.
        pod.podcache.dependency_graph.add(doc.pod_path, doc.pod_path)
        def _track_dep(path):
            pod.podcache.dependency_graph.add(doc.pod_path, path)
        tracking_func = _track_dep

    base_loader = make_base_yaml_loader(
        pod, locale=locale, untag_params=untag_params,
        tracking_func=tracking_func)

    class YamlLoader(base_loader):

        @staticmethod
        def loader_locale():
            return str(doc.locale_safe) if doc else locale

        @staticmethod
        def pod_path():
            if doc:
                return doc.pod_path
            return None

        def construct_doc(self, node):
            def func(path):
                constructed_doc = pod.get_doc(path, locale=self.loader_locale())
                if not constructed_doc.exists:
                    raise errors.DocumentDoesNotExistError(
                        'Referenced document does not exist: {}'.format(path))
                tracking_func(constructed_doc.pod_path)
                return constructed_doc
            return self._construct_func(node, func)

        def construct_static(self, node):
            def func(path):
                tracking_func(path)
                return pod.get_static(path, locale=self.loader_locale())
            return self._construct_func(node, func)

        def construct_url(self, node):
            def func(path):
                tracking_func(path)
                return pod.get_url(path, locale=self.loader_locale())
            return self._construct_func(node, func)

    YamlLoader.add_constructor('!g.doc', YamlLoader.construct_doc)
    YamlLoader.add_constructor('!g.static', YamlLoader.construct_static)
    YamlLoader.add_constructor('!g.url', YamlLoader.construct_url)

    return YamlLoader


def load_yaml(*args, **kwargs):
    pod = kwargs.pop('pod', None)
    doc = kwargs.pop('doc', None)
    untag_params = kwargs.pop('untag_params', None)
    simple_loader = kwargs.pop('simple_loader', False)
    default_locale = None
    if doc:
        default_locale = doc._locale_kwarg or doc.collection.default_locale
    locale = kwargs.pop('locale', default_locale)
    if simple_loader:
        # Simple loader does not reference collections or documents.
        loader = make_base_yaml_loader(
            pod, locale=locale, untag_params=untag_params)
    else:
        loader = make_yaml_loader(
            pod, doc=doc, locale=locale, untag_params=untag_params)
    contents = yaml.load(*args, Loader=loader, **kwargs) or {}
    if not untag_params:
        return contents
    return untag.Untag.untag(
        contents, locale_identifier=locale, params=untag_params)


def load_plain_yaml(content, pod=None, locale=None):
    return yaml.load(content, Loader=yaml_utils.PlainTextYamlLoader)


def parse_yaml(content, pod=None, locale=None, untag_params=None,
               simple_loader=False):
    return load_yaml(
        content, pod=pod, locale=locale, untag_params=untag_params,
        simple_loader=simple_loader)


def dump_yaml(obj):
    """Dumps yaml using the the safe dump."""
    return yaml.safe_dump(
        obj, allow_unicode=True, width=800, default_flow_style=False)


def dump_plain_yaml(obj):
    """Dumps yaml using the plain text dumper to retain constructors."""
    return yaml.dump(
        obj, Dumper=yaml_utils.PlainTextYamlDumper,
        default_flow_style=False, allow_unicode=True, width=800)


def ordered_dict_representer(dumper, data):
    return dumper.represent_mapping('tag:yaml.org,2002:map', list(data.items()))


yaml.SafeDumper.add_representer(OrderedDict, ordered_dict_representer)


def slugify(text, delim='-'):
    if not isinstance(text, str):
        text = str(text)
    result = []
    for word in SLUG_REGEX.split(text.lower()):
        if not isinstance(word, str):
            word = word
        word = codecs.encode(word, 'translit/long')
        if word:
            result.append(word)
    slug = str(delim.join(result))
    for seq, sub in SLUG_SUBSTITUTE:
        slug = slug.replace(seq.format(delim), sub.format(delim))
    return slug


class DummyDict(object):

    def __getattr__(self, name):
        return ''


class JsonEncoder(json.JSONEncoder):

    # pylint: disable=method-hidden
    def default(self, o):
        if hasattr(o, 'timetuple'):
            return time.mktime(o.timetuple())
        raise TypeError(repr(o) + ' is not JSON serializable.')


def LocaleIterator(iterator, locale):
    locale = str(locale)
    for i, line in enumerate(iterator):
        if i == 0 or line.startswith(locale):
            yield line


def get_rows_from_csv(pod, path, locale=SENTINEL):
    fp = pod.open_file(path)
    if locale is not SENTINEL:
        fp = LocaleIterator(fp, locale=locale)
    rows = []
    for row in csv_lib.DictReader(fp):
        data = {}
        for header, cell in row.items():
            if cell is None:
                cell = ''
            data[header] = cell
        rows.append(data)
    return rows


class ProgressBarThread(threading.Thread):

    def __init__(self, bar, enabled, *args, **kwargs):
        self.bar = bar
        self.enabled = enabled
        super(ProgressBarThread, self).__init__(*args, **kwargs)

    def run(self):
        super(ProgressBarThread, self).run()
        if self.enabled:
            self.bar.update(self.bar.value + 1)


def clean_html(content, convert_to_markdown=False):
    soup = bs4.BeautifulSoup(content, 'html.parser')
    _process_google_hrefs(soup)
    _process_google_comments(soup)
    # Support HTML fragments without body tags.
    content = str(soup.body or soup)
    if convert_to_markdown:
        h2t = html2text.HTML2Text()
        h2t.body_width = 0  # https://github.com/grow/grow/issues/887
        # Fixes a bug with malformed tables from Google Doc integration.
        # Preserves tables as <table> tags.
        h2t.bypass_tables = True
        h2t.single_line_break = False
        content = h2t.handle(content).strip()
    return content


def _process_google_hrefs(soup):
    for tag in soup.find_all('a'):
        if tag.attrs.get('href'):
            tag['href'] = _clean_google_href(tag['href'])


def _process_google_comments(soup):
    for el in soup.find_all('a'):
        id_attribute = el.get('id')
        if id_attribute and id_attribute.startswith(('cmnt', 'ftnt')):
            footer_parent = el.find_parent('div')
            if footer_parent:
                footer_parent.decompose()
            sup_parent = el.find_parent('sup')
            if sup_parent:
                sup_parent.decompose()


def _clean_google_href(href):
    regex = ('^'
             + re.escape('https://www.google.com/url')
             + '(\?|.*\&)q=([^\&]*)')
    match = re.match(regex, href)
    if match:
        encoded_url = match.group(2)
        return url_parse.unquote(encoded_url)
    return href


def format_existing_data(old_data, new_data, preserve=None, key_to_update=None):
    if old_data:
        if preserve == 'builtins':
            for key in list(old_data.keys()):
                if not key.startswith('$'):
                    del old_data[key]
        if key_to_update:
            old_data[key_to_update] = new_data
        else:
            old_data.update(new_data)
        return old_data
    if key_to_update:
        return {key_to_update: new_data}
    return new_data


def fnmatches_paths(path_to_extract, paths):
    # Special case: user doesn't want to check against any paths.
    if not paths:
        return True
    for path in paths:
        # Support pod paths and filesystem paths for tab completion.
        path_to_extract = path_to_extract.lstrip('/')
        path = path.lstrip('/')
        matched = fnmatch.fnmatch(path_to_extract, path)
        if matched:
            return True
    return False



"""Preprocessor for translations."""

import re
from . import base

SUFFIXES = frozenset(['po'])
SUFFIX_PATTERN = re.compile('[.](' + '|'.join(map(re.escape, SUFFIXES)) + ')$')


class TranslationPreprocessor(base.BasePreprocessor):
    """Automatically recompiles translations when needed."""
    KIND = '_translation'

    def __init__(self, pod):
        self.pod = pod

    def run(self, build=True):
        self.pod.catalogs.compile()

    def list_watched_dirs(self):
        return ['/translations/']



from .destinations import messages
import logging
import texttable


def print_results(results):
    for message in results.test_results:
        if message.result == messages.Result.PASS:
            color = texttable.bcolors.GREEN
        elif message.result == messages.Result.FAIL:
            color = texttable.bcolors.GREEN
        else:
            color = texttable.bcolors.YELLOW
        label = texttable.get_color_string(color, message.result)
        logging.info('{} {}'.format(label, message.title))



"""Grow local development server."""

import os
import pkg_resources
from urllib import parse as url_parse
from werkzeug import serving
from werkzeug.middleware import shared_data
from grow.common import utils
from grow.pods import errors
from grow.sdk import sdk_utils
from grow.server import handlers


# Use grow's logger instead of werkzeug's default.
class RequestHandler(serving.WSGIRequestHandler):

    @property
    def server_version(self):
        version = pkg_resources.get_distribution('grow').version
        return 'Grow/{}'.format(version)

    def log(self, *args, **kwargs):
        pass


class PodServer(object):

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def __init__(self, pod, host, port, debug=False):
        self.pod = pod
        self.host = host
        self.port = port
        self.pod.render_pool.pool_size = 1
        self.debug = debug
        self.routes = self.pod.router.routes

        # Trigger the dev handler hook.
        self.pod.extensions_controller.trigger(
            'dev_handler', self.routes, host=host, port=port, debug=debug)

        # Start off the server with a clean dependency graph.
        self.pod.podcache.dependency_graph.mark_clean()

    def dispatch_request(self, request):
        path = url_parse.unquote(request.path)  # Support escaped paths.
        matched = self.routes.match(path)

        if not matched:
            text = '{} was not found in routes.'
            raise errors.RouteNotFoundError(text.format(path))

        kind = matched.value.kind
        if kind == 'console':
            if 'handler' in matched.value.meta:
                handler_meta = None
                if 'meta' in matched.value.meta:
                    handler_meta = matched.value.meta['meta']
                return matched.value.meta['handler'](
                    self.pod, request, matched, meta=handler_meta)
            return handlers.serve_console(self.pod, request, matched)
        return handlers.serve_pod(self.pod, request, matched)

    def wsgi_app(self, environ, start_response):
        try:
            request = handlers.Request(environ)
            response = self.dispatch_request(request)
            return response(environ, start_response)
        except Exception as e:
            request = handlers.Request(environ)
            response = handlers.serve_exception(self.pod, request, e)
            return response(environ, start_response)


def create_wsgi_app(pod, host, port, debug=False):
    podserver_app = PodServer(pod, host, port, debug=debug)
    assets_path = os.path.join(utils.get_grow_dir(), 'ui', 'admin', 'assets')
    ui_path = os.path.join(utils.get_grow_dir(), 'ui', 'dist')
    # pylint: disable=no-member
    return shared_data.SharedDataMiddleware(podserver_app, {
        '/_grow/ui': ui_path,
        '/_grow/assets': assets_path,
    })



"""Translation stats collection and reporting."""

import datetime
import io
import logging
import traceback
import texttable


class TranslationStats(object):

    ROW_COUNT = 7

    def __init__(self):
        self._locale_to_message = {}
        self._untranslated = {}
        self._stacktraces = []
        self._untagged = {}
        # Default to info logging.
        self.log = logging.info
        self.datetime = datetime.datetime

    @property
    def messages(self):
        """All messages and counts."""
        return self._locale_to_message

    @property
    def missing(self):
        """Messages that are untagged and untranslated during rendering."""
        untagged_messages = set([msg for _, msg in self.untagged.items()])
        tracking = {}
        for locale, messages in self._untranslated.items():
            if locale not in tracking:
                tracking[locale] = {}
            for message in messages:
                if message in untagged_messages:
                    tracking[locale][message] = self._locale_to_message[
                        locale][message]
        # Removing empty locales.
        blank_locales = []
        for key in tracking:
            if not tracking[key]:
                blank_locales.append(key)
        for key in blank_locales:
            del tracking[key]
        return tracking

    @property
    def untagged(self):
        """Untagged messages by location."""
        return self._untagged

    @property
    def untranslated(self):
        """Untranslated messages by locale."""
        tracking = {}
        for locale, messages in self._untranslated.items():
            if locale not in tracking:
                tracking[locale] = {}
            for message in messages:
                tracking[locale][message] = self._locale_to_message[
                    locale][message]
        return tracking

    @property
    def count_untranslated(self):
        """Max number of untranslated strings for any locale."""
        untranslated_ids = set()
        for _, messages in self._untranslated.items():
            for message in messages:
                untranslated_ids.add(message)
        return len(untranslated_ids)

    @property
    def stacktraces(self):
        return self._stacktraces

    @staticmethod
    def _simplify_traceback(tb):
        """Remove extra layers of the traceback to just get best bits."""
        start = 0
        end = len(tb) - 3
        for i, item in enumerate(tb):
            if 'jinja' in item:
                start = i + 1
                break
        return tb[start:end]

    def _get_message_locations(self, locale, message):
        """Get a list of the locations for a given locale and message."""
        locations = set()

        for item in self.stacktraces:
            if item['locale'] != locale or item['id'] != message:
                continue
            if item['location']:
                locations.add(item['location'])

        return [(location, None) for location in locations]

    def add_untagged(self, paths_to_untagged):
        """Add untagged paths and strings."""
        self._untagged = paths_to_untagged

    def export(self):
        """Export messages and untranslated strings."""
        return {
            'messages': self.messages,
            'untranslated': self.untranslated,
            'untagged': self.untagged,
        }

    def export_untranslated_catalogs(self, pod, dir_path=None):
        """Export the untranslated messages into catalogs based on locale."""
        locale_to_catalog = {}
        for locale, messages in self.untranslated.items():
            if locale not in locale_to_catalog:
                locale_to_catalog[locale] = pod.catalogs.get(
                    locale, basename='untranslated.po', dir_path=dir_path)
            catalog = locale_to_catalog[locale]
            for message in messages:
                catalog.add(message, locations=self._get_message_locations(
                    locale, message))
        return locale_to_catalog

    def export_untranslated_tracebacks(self):
        """Export the untranslated tracebacks into a log file."""
        width = 80
        border_width = 3
        border_char = '='

        def _blank_line():
            return '\n'

        def _solid_line():
            return '{}\n'.format(border_char * width)

        def _text_line(text):
            text_width = width - (border_width + 1) * 2
            return '{} {} {}\n'.format(
                border_char * border_width, text.center(text_width),
                border_char * border_width)

        with io.StringIO() as output:
            output.write(_solid_line())
            output.write(_text_line('Untranslated Strings'))
            output.write(_solid_line())
            output.write(_text_line(
                '{} occurrences of {} untranslated strings'.format(
                    len(self.stacktraces), self.count_untranslated)))
            output.write(_text_line(str(self.datetime.now())))
            output.write(_solid_line())
            output.write(_blank_line())

            if not self.stacktraces:
                output.write(
                    _text_line('No untranslated strings found.'))
                return output.getvalue()

            for item in self.stacktraces:
                output.write('{} :: {}\n'.format(item['locale'], item['id']))
                for line in item['tb']:
                    output.write('{}'.format(line))
                output.write(_blank_line())

            return output.getvalue()

    def tick(self, message, locale, default_locale, location=None):
        """Count a translation."""
        if not message:
            return

        if locale not in self._locale_to_message:
            self._locale_to_message[locale] = {}

        messages = self._locale_to_message[locale]
        if message.id not in messages:
            messages[message.id] = 0
        messages[message.id] += 1

        # Check for untranslated message.
        if not message.string and message.id.strip() and locale is not default_locale:
            if locale not in self._untranslated:
                self._untranslated[locale] = set()

            stack = traceback.format_stack()
            self._stacktraces.append({
                'locale': locale,
                'location': location,
                'id': message.id,
                'tb': self._simplify_traceback(stack),
            })
            self._untranslated[locale].add(message.id)

    def pretty_print(self, show_all=False):
        """Outputs the translation stats to a table formatted view."""

        if not self.untranslated and not self.untagged:
            self.log('\nNo untranslated strings found.\n')
            return

        # Most frequent untranslated and untagged messages.
        if self.untagged:
            table = texttable.Texttable(max_width=120)
            table.set_deco(texttable.Texttable.HEADER)
            table.set_cols_dtype(['t', 'i', 't'])
            table.set_cols_align(['l', 'r', 'l'])
            rows = []

            missing = self.missing
            for locale in missing:
                for message in missing[locale]:
                    rows.append([str(locale), missing[locale][message], message])

            rows = sorted(rows, key=lambda x: -x[1])
            if not show_all:
                num_rows = len(rows)
                rows = rows[:self.ROW_COUNT]
                if num_rows > self.ROW_COUNT:
                    rows.append(['', num_rows - self.ROW_COUNT,
                                 '+ Additional untranslated strings...'])

            table.add_rows(
                [['Locale', '#', 'Untagged and Untranslated Message']] + rows)
            self.log('\n' + table.draw() + '\n')

        # Most frequent untranslated messages.
        if self.untranslated:
            table = texttable.Texttable(max_width=120)
            table.set_deco(texttable.Texttable.HEADER)
            table.set_cols_dtype(['t', 'i', 't'])
            table.set_cols_align(['l', 'r', 'l'])
            rows = []

            for locale in self.untranslated:
                for message in self.untranslated[locale]:
                    rows.append([str(locale), self.untranslated[
                                locale][message], message])

            rows = sorted(rows, key=lambda x: -x[1])
            if not show_all:
                num_rows = len(rows)
                rows = rows[:self.ROW_COUNT]
                if num_rows > self.ROW_COUNT:
                    rows.append(['', num_rows - self.ROW_COUNT,
                                 '+ Additional untranslated strings...'])

            table.add_rows([['Locale', '#', 'Untranslated Message']] + rows)
            self.log('\n' + table.draw() + '\n')

            # Untranslated messages per locale.
            table = texttable.Texttable(max_width=120)
            table.set_deco(texttable.Texttable.HEADER)
            table.set_cols_dtype(['t', 'i'])
            table.set_cols_align(['l', 'r'])
            rows = []

            for locale in self.untranslated:
                rows.append([str(locale), len(self.untranslated[locale])])

            rows = sorted(rows, key=lambda x: -x[1])

            table.add_rows([['Locale', 'Untranslated']] + rows)
            self.log('\n' + table.draw() + '\n')



"""Base class for the post render hook."""

from grow.extensions.hooks import base_hook


class PostRenderHook(base_hook.BaseHook):
    """Hook for post render."""

    KEY = 'post_render'
    NAME = 'Post Render'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, doc, raw_content, *_args, **_kwargs):
        """Trigger the post render hook."""
        if previous_result:
            return previous_result
        return raw_content



"""Tracks the rendered document information."""

import hashlib
import os


class RenderedDocument(object):
    """Keeps track of the information for the rendered document."""

    def __init__(self, path, content=None, tmp_dir=None):
        self.path = path
        self.tmp_dir = tmp_dir
        self.hash = None
        self._content = None

        # When doing threaded rendering the thread cannot update the timer.
        # Keep the timer with the rendered document to add to the normal
        # timers.
        self.render_timer = None
        self.write(content)

    def _has_tmp_file_path(self):
        return self.tmp_dir is not None and self.hash is not None

    def _get_tmp_file_path(self):
        """Temp filename if has temporary dir."""
        return os.path.join(self.tmp_dir, self.hash)

    @property
    def file_path(self):
        """Returns the temp file name if available."""
        if not self._has_tmp_file_path():
            return None
        return self._get_tmp_file_path()

    def read(self):
        """Reads the content when it needs it."""
        if not self._has_tmp_file_path():
            return self._content

        with open(self._get_tmp_file_path(), "r") as tmp_file:
            file_contents = tmp_file.read()
            return file_contents

    def write(self, content):
        """Writes the content to the temp filesystem or keeps in memory."""
        if content is None:
            self.hash = None
        else:
            try:
                self.hash = hashlib.sha1(content.encode('utf-8')).hexdigest()
            except AttributeError:
                self.hash = hashlib.sha1(content).hexdigest()

        if self._has_tmp_file_path():
            with open(self._get_tmp_file_path(), "w") as tmp_file:
                tmp_file.write(content)
        else:
            self._content = content



"""Renderer for performing render operations for the pod."""

import progressbar
from grow.common import progressbar_non
from grow.common import bulk_errors
from grow.rendering import render_batch


class Renderer(object):
    """Handles the rendering and threading of the controllers."""

    @staticmethod
    def rendered_docs(pod, routes, use_threading=True, source_dir=None):
        """Generate the rendered documents for the given routes."""
        with pod.profile.timer('renderer.Renderer.render_docs'):
            routes_len = len(routes)
            text = 'Building: %(value)d/{} (in %(time_elapsed).9s)'
            widgets = [progressbar.FormatLabel(text.format(routes_len))]
            progress = progressbar_non.create_progressbar(
                "Building pod...", widgets=widgets, max_value=routes_len)
            progress.start()

            def tick():
                """Tick the progress bar value."""
                progress.update(progress.value + 1)

            batches = render_batch.RenderBatches(
                pod.render_pool, pod.profile, tick=tick)

            for controller in Renderer.controller_generator(pod, routes):
                batches.add(controller)

            if source_dir:
                # When using an input directory, load the files instead of render.
                rendered_docs, render_errors = batches.load(
                    use_threading=use_threading, source_dir=source_dir)
            else:
                # Default to rendering the documents.
                rendered_docs, render_errors = batches.render(
                    use_threading=use_threading)

            progress.finish()

            if render_errors:
                text = 'There were {} errors during rendering.'
                raise bulk_errors.BulkErrors(text.format(
                    len(render_errors)), render_errors)
            return rendered_docs

    @staticmethod
    def controller_generator(pod, routes):
        """Generate the controllers for the given routes."""
        for path, route_info, _ in routes.nodes:
            yield pod.router.get_render_controller(path, route_info)



"""Base class for the preprocess hook."""

import json
from protorpc import messages
from protorpc import protojson
from grow.extensions.hooks import base_hook


class PreProcessHook(base_hook.BaseHook):
    """Hook for preprocess."""

    KEY = 'preprocess'
    NAME = 'Preprocess'
    KIND = None

    class Config(messages.Message):
        """Config for preprocessing."""
        pass

    @classmethod
    def parse_config(cls, config):
        """Parse the config into the protobuf."""
        return protojson.decode_message(
            cls.Config, json.dumps(config))

    # pylint: disable=arguments-differ, unused-argument
    def should_trigger(self, previous_result, config, names, tags, run_all, *_args, **_kwargs):
        """Determine if the preprocess should trigger."""
        name = config.get('name')
        kind = config.get('kind')
        config_tags = config.get('tags', [])
        autorun = config.get('autorun', True)

        if kind != self.KIND:
            return False

        if names and name in names:
            return True
        elif tags and set(config_tags).intersection(tags):
            return True
        elif autorun or run_all:
            return True
        return False

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, config, names, tags, run_all, rate_limit, *_args, **_kwargs):
        """Trigger the preprocess hook."""
        if previous_result:
            return previous_result
        return None



"""Stub hook for fallback when older versions don't support newer hooks."""

from grow.extensions.hooks import base_hook


class StubHook(base_hook.BaseHook):
    """Hook for stubbing for newer hooks."""

    KEY = 'stub'
    NAME = 'Stub Handler'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, *_args, **_kwargs):
        """Trigger the dev handler hook."""
        return previous_result



"""Custom yaml loaders for Grow."""

import datetime
import types
from collections import OrderedDict
import yaml
from yaml.representer import Representer
from yaml.representer import SafeRepresenter

try:
    from yaml import CLoader as yaml_Loader
except ImportError:
    from yaml import Loader as yaml_Loader


class PlainText(dict):

    def __init__(self, tag, value):
        super(PlainText, self)
        self['tag'] = tag
        self['value'] = value


class PlainTextYamlLoader(yaml_Loader):
    """Loader for loading as plain text."""

    def construct_plaintext(self, node):
        return PlainText(node.tag, node.value)


class PlainTextYamlDumper(yaml.Dumper):
    """Dumper for working with plain text yaml."""
    pass


def plain_text_dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


def plain_text_dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def plain_text_representer(dumper, data):
    return dumper.represent_scalar(data['tag'], data['value'])


# Don't want to actually process the constructors, just keep the values
PlainTextYamlDumper.add_representer(OrderedDict, plain_text_dict_representer)
PlainTextYamlDumper.add_representer(PlainText, plain_text_representer)
PlainTextYamlDumper.add_representer(type(None), SafeRepresenter.represent_none)
PlainTextYamlDumper.add_representer(str, SafeRepresenter.represent_str)
PlainTextYamlDumper.add_representer(bool, SafeRepresenter.represent_bool)
PlainTextYamlDumper.add_representer(int, SafeRepresenter.represent_int)
PlainTextYamlDumper.add_representer(float, SafeRepresenter.represent_float)
PlainTextYamlDumper.add_representer(list, SafeRepresenter.represent_list)
PlainTextYamlDumper.add_representer(tuple, SafeRepresenter.represent_list)
PlainTextYamlDumper.add_representer(dict, SafeRepresenter.represent_dict)
PlainTextYamlDumper.add_representer(set, SafeRepresenter.represent_set)
PlainTextYamlDumper.add_representer(
    datetime.date, SafeRepresenter.represent_date)
PlainTextYamlDumper.add_representer(
    datetime.datetime, SafeRepresenter.represent_datetime)
PlainTextYamlDumper.add_representer(None, SafeRepresenter.represent_undefined)
PlainTextYamlDumper.add_multi_representer(object, Representer.represent_object)


PlainTextYamlLoader.add_constructor(
    '!_', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.csv', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.doc', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.json', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.static', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.string', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.url', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.yaml', PlainTextYamlLoader.construct_plaintext)



"""Filter paths to control ignores and includes by path."""

import re


DEFAULT_IGNORED = [
    re.compile(r'^.*/\.[^/]*$'),  # Dot files.
]
DEFAULT_INCLUDED = []


class PathFilter(object):
    """Filter for testing paths against a set of filter criteria."""

    def __init__(self, ignored=None, included=None):
        self._ignored = []
        self._included = []

        if ignored:
            for item in ignored:
                self.add_ignored(item)

        if included:
            for item in included:
                self.add_included(item)

    def __repr__(self):
        return '<PathFilter ignored={} included={}>'.format(
            [i.pattern for i in self.ignored], [i.pattern for i in self.included])

    def _is_ignored(self, path):
        """Test for ignored pattern match."""
        for pattern in self.ignored:
            if pattern.search(path):
                return True
        return False

    def _is_included(self, path):
        """Test for include pattern match."""
        for pattern in self.included:
            if pattern.search(path):
                return True
        return False

    @property
    def ignored(self):
        """All ignored patterns."""
        if not self._ignored and DEFAULT_IGNORED:
            self._ignored = DEFAULT_IGNORED
        for item in self._ignored:
            yield item

    @property
    def included(self):
        """All included patterns."""
        if not self._included and DEFAULT_INCLUDED:
            self._included = DEFAULT_INCLUDED
        for item in self._included:
            yield item

    def add_ignored(self, raw_pattern):
        """Add a new ignored pattern."""
        self._ignored.append(re.compile(raw_pattern))

    def add_included(self, raw_pattern):
        """Add a new included pattern."""
        self._included.append(re.compile(raw_pattern))

    def export(self):
        """Export for serialization."""
        return {
            'ignored': [item.pattern for item in self._ignored],
            'included': [item.pattern for item in self._included],
        }

    def is_valid(self, path):
        """Tests if the path is valid according to the known filters."""
        if self._is_ignored(path):
            # Includes override an ignores.
            return self._is_included(path)
        return True



"""Grow UI Tools"""

import os
import jinja2
from grow.common import utils
from grow import storage
from grow.templates import filters


@utils.memoize
def create_jinja_env():
    root = os.path.join(utils.get_grow_dir(), 'ui', 'admin')
    loader = storage.FileStorage.JinjaLoader(root)
    env = jinja2.Environment(
        loader=loader,
        autoescape=True,
        trim_blocks=True,
        extensions=[
            'jinja2.ext.autoescape',
            'jinja2.ext.do',
            'jinja2.ext.i18n',
            'jinja2.ext.loopcontrols',
            'jinja2.ext.with_',
        ])
    env.filters.update(filters.create_builtin_filters(env, None))
    return env



"""
Cache for storing and retrieving file content at a pod_path.

The contents of the cache can be parsed and are stored based on locale.
"""

class FileCache(object):
    """Simple cache for file contents."""

    def __init__(self):
        self.reset()

    def _ensure_exists(self, pod_path, value=None):
        if pod_path not in self._cache:
            self._cache[pod_path] = value or {}
        return self._cache[pod_path]

    def add(self, pod_path, value, locale=None):
        """Add a value to the cache by pod_path."""
        container = self._ensure_exists(pod_path)
        container[locale] = value

    def add_all(self, pod_path_to_cached, locale=None):
        """Add a multiple values to the cache by pod_paths."""
        for pod_path, value in pod_path_to_cached.items():
            container = self._ensure_exists(pod_path)
            container[locale] = value

    def remove(self, pod_path):
        """Remove a value from the cache by pod_path."""
        return self._cache.pop(pod_path, None)

    def export(self):
        """Exports all the file cache data."""
        return self._cache

    def get(self, pod_path, locale=None):
        """Retrieve a cache value by pod_path or None."""
        container = self._ensure_exists(pod_path)
        return container.get(locale, None)

    def reset(self):
        """Resets the internal cache reference."""
        self._cache = {}



"""Partials contain organized units for templating."""

import os
from grow.partials import partial as grow_partial


class Partials(object):
    """Manage partials within the pod."""

    PARTIALS_PATH = '/partials'
    VIEWS_PATH = '/views/partials'

    def __init__(self, pod):
        self.pod = pod

    def __repr__(self):
        return '<Partials "{}">'.format(self.PARTIALS_PATH)

    def get_partial(self, key):
        """Returns a specific partial in the pod."""
        # Check if the partial is a view partial.
        pod_path = '{}/{}.html'.format(self.VIEWS_PATH, key)
        if self.pod.file_exists(pod_path):
            return grow_partial.ViewPartial(pod_path, self.pod)

        pod_path = '{}/{}'.format(self.PARTIALS_PATH, key)
        return grow_partial.Partial(pod_path, self.pod)

    def get_partials(self):
        """Returns all partials in the pod."""
        items = []

        # Load all partials in the partials directory.
        abs_base_dir = self.pod.abs_path(self.PARTIALS_PATH)
        for root, dirs, _ in self.pod.walk(self.PARTIALS_PATH):
            if root == abs_base_dir:
                for dir_name in dirs:
                    pod_path = '{}/{}'.format(self.PARTIALS_PATH, dir_name)
                    partial = grow_partial.Partial(pod_path, self.pod)
                    items.append(partial)

        # Load all partials in the views partials directory.
        view_pod_paths = []
        for root, dirs, files in self.pod.walk(self.VIEWS_PATH):
            pod_dir = root.replace(self.pod.root, '')
            for file_name in files:
                pod_path = os.path.join(pod_dir, file_name)
                partial = grow_partial.ViewPartial(pod_path, self.pod)
                items.append(partial)

        return items



"""RC Config control class."""

import datetime
import os
import time
import yaml
from grow.common import base_config


RC_FILE_NAME = '.growrc.yaml'
RC_LAST_CHECKED_DELTA = datetime.timedelta(hours=1)


class RCConfig(base_config.BaseConfig):
    """Config for Grow RC file."""

    def __init__(self, config=None, internal_time=time.time):
        super(RCConfig, self).__init__(config=config)
        self._time = internal_time
        # Working directory default config.
        self._wd_config = base_config.BaseConfig()
        if config is None:
            self.read()

    @staticmethod
    def _is_ci_env():
        if 'CI' in os.environ:
            return True
        return False

    @property
    def filename(self):
        """Filename of the RC File."""
        return os.path.expanduser('~/{}'.format(RC_FILE_NAME))

    @property
    def last_checked(self):
        """Timestamp of the last time checked for sdk update."""
        return self.get('update.last_checked', 0)

    @last_checked.setter
    def last_checked(self, value):
        """Timestamp of the last time checked for sdk update."""
        return self.set('update.last_checked', value)

    @property
    def needs_update_check(self):
        """Check if the update check needs to be done."""
        if self._is_ci_env():
            return False
        time_passed = self._time() - self.last_checked
        return time_passed > RC_LAST_CHECKED_DELTA.total_seconds()

    def get(self, identifier, default_value=None):
        """Retrieve the identifier value in the config after checking local config."""
        base_default = super(RCConfig, self).get(identifier, default_value)
        return self._wd_config.get(identifier, default_value=base_default)

    def read(self):
        """Reads the RC config from the system."""
        # Read the config from the home directory.
        filename = self.filename
        if not os.path.isfile(filename):
            self._config = {}
        else:
            with open(filename, 'r') as conf:
                self._config = yaml.safe_load(conf.read()) or {}

        # Allow for an read only override from the working directory rc file.
        wd_filename = os.path.join(os.getcwd(), RC_FILE_NAME)
        if os.path.isfile(wd_filename):
            with open(wd_filename, 'r') as conf:
                self._wd_config = base_config.BaseConfig(
                    yaml.safe_load(conf.read()))

    def reset_update_check(self):
        """Reset the timestamp of the last_checked."""
        self.last_checked = self._time()

    def write(self):
        """Writes the RC config to the system."""
        rc_file_name = self.filename
        with open(rc_file_name, 'w') as conf:
            conf.write(yaml.safe_dump(self._config, default_flow_style=False))


# Global config for working with the rc config without needing to re-read
# the config.
RC_CONFIG = RCConfig()



"""A pod encapsulates all files used to build a site."""

import copy
import atexit
import json
import logging
import os
import sys
import threading
import time
import shutil
import tempfile
import yaml
import jinja2
import progressbar
import cachelib
from grow import storage as grow_storage
from grow.cache import podcache
from grow.collections import collection
from grow.common import deprecated
from grow.common import features
from grow.common import logger
from grow.common import progressbar_non
from grow.common import untag
from grow.common import utils
from grow.documents import static_document
from grow.extensions import extension_controller as ext_controller
from grow.extensions import extension_importer
from grow.partials import partials
from grow.performance import profile
from grow.pods import errors
from grow.preprocessors import preprocessors
from grow.rendering import markdown_utils
from grow.rendering import rendered_document
from grow.rendering import renderer
from grow.rendering import render_pool as grow_render_pool
from grow.routing import path_filter as grow_path_filter
from grow.routing import path_format as grow_path_format
from grow.routing import router as grow_router
from grow.templates import filters
from grow.templates import jinja_dependency
from grow.templates import tags
from grow.translations import catalog_holder
from grow.translations import locales
from grow.translations import translation_stats
from grow.translators import translators
from . import env as environment
from . import podspec


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class PodDoesNotExistError(Error, IOError):
    pass


# Pods can create temp directories. Need to track temp dirs for cleanup.
_POD_TEMP_DIRS = []


@atexit.register
def goodbye_pods():
    for tmp_dir in _POD_TEMP_DIRS:
        shutil.rmtree(tmp_dir, ignore_errors=True)


class Pod(object):
    """Grow pod."""
    INSTALLED_EXTENSIONS_DIR_PATH = 'extensions'
    LOCAL_EXTENSIONS_DIR_PATH = 'ext'
    FEATURE_UI = 'ui'
    FEATURE_TRANSLATION_STATS = 'translation_stats'
    FEATURE_OLD_SLUGIFY = 'legacy_slugify'
    FILE_DEP_CACHE = 'depcache.json'
    FILE_PODSPEC = 'podspec.yaml'
    FILE_EXTENSIONS = 'extensions.txt'
    PATH_CONTROL = '/.grow/'

    def __eq__(self, other):
        return (isinstance(self, Pod)
                and isinstance(other, Pod)
                and self.root == other.root)

    def __init__(self, root, storage=grow_storage.AUTO, env=None, load_extensions=True):
        self.deprecated = deprecated.DeprecationManager(self.logger.warn)
        self._yaml = utils.SENTINEL
        self.storage = storage
        self.root = (root if self.storage.is_cloud_storage
                     else os.path.abspath(root))
        self.env = (env if env
                    else environment.Env(environment.EnvConfig(host='localhost')))
        self.locales = locales.Locales(pod=self)
        self.catalogs = catalog_holder.Catalogs(pod=self)
        self._jinja_env_lock = threading.RLock()
        self._podcache = None
        self._features = features.Features(disabled=[
            self.FEATURE_TRANSLATION_STATS,
            self.FEATURE_OLD_SLUGIFY,
        ])
        self._experiments = features.Features(default_enabled=False)

        self._extensions_controller = ext_controller.ExtensionController(self)

        if self.exists:
            # Modify sys.path for built-in extension support.
            _installed_ext_dir = self.abs_path(self.extensions_dir)
            if os.path.exists(_installed_ext_dir):
                sys.path.insert(0, _installed_ext_dir)

            # Modify sys.path for local extension support.
            _local_ext_dir = self.abs_path(Pod.LOCAL_EXTENSIONS_DIR_PATH)
            if os.path.exists(_local_ext_dir):
                sys.path.insert(0, _local_ext_dir)

            # Load the features from the podspec.
            self._load_features()

            # Load the experiments from the podspec.
            self._load_experiments()

        # Ensure preprocessors are loaded when pod is initialized.
        # Preprocessors may modify the environment in ways that are required by
        # data files (e.g. yaml constructors). Avoid loading extensions using
        # `load_extensions=False` to permit `grow install` to be used to
        # actually install extensions, prior to loading them.
        if load_extensions and self.exists:
            self.list_preprocessors()

        # Load extensions, ignore local extensions during install.
        self._load_extensions(load_extensions)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '<Pod: {}>'.format(self.root)

    # DEPRECATED: Remove when no longer needed after re-route stable release.
    def _get_bytecode_cache(self):
        return self.jinja_bytecode_cache

    def _load_experiments(self):
        config = self.yaml.get('experiments', {})
        for key, value in config.items():
            # Expertiments can be turned on with a True value,
            # be turned off with a False value,
            # or turned on by a providing configuration value.
            if value is True:
                self._experiments.enable(key)
            elif value is False:
                self._experiments.disable(key)
            else:
                self._experiments.enable(key, config=value)

    def _load_extensions(self, load_local_extensions=True):
        self._extensions_controller.register_builtins()

        if load_local_extensions and self.exists:
            self._extensions_controller.register_extensions(
                self.yaml.get('ext', []))

    def _load_features(self):
        config = self.yaml.get('features', {})
        for key, value in config.items():
            # Features can be turned on with a True value,
            # be turned off with a False value,
            # or turned on by a providing configuration value.
            if value is True:
                self._features.enable(key)
            elif value is False:
                self._features.disable(key)
            else:
                self._features.enable(key, config=value)

    def _normalize_path(self, pod_path):
        if '..' in pod_path:
            raise ValueError('.. not allowed in file paths.')
        return os.path.join(self.root, pod_path.lstrip('/'))

    def _normalize_pod_path(self, pod_path):
        if '..' in pod_path:
            raise ValueError('.. not allowed in pod paths.')
        if not pod_path.startswith('/'):
            pod_path = '/{}'.format(pod_path)
        return pod_path

    def _parse_object_cache_file(self):
        with self.profile.timer('Pod._parse_object_cache_file'):
            object_cache_file_name = '/{}'.format(podcache.FILE_OBJECT_CACHE)

            if not self.file_exists(object_cache_file_name):
                return {}
            try:
                return self.read_json(object_cache_file_name) or {}
            except IOError:
                path = self.abs_path(object_cache_file_name)
                raise podcache.PodCacheParseError(
                    'Error parsing: {}'.format(path))

    def _parse_dep_cache_file(self):
        with self.profile.timer('Pod._parse_dep_cache_file'):
            podcache_file_name = '/{}'.format(self.FILE_DEP_CACHE)

            # TODO Remove deprecated cachefile support.
            # Convert legacy yaml cache files.
            legacy_podcache_file_name = '/.podcache.yaml'
            if self.file_exists(legacy_podcache_file_name):
                logging.info(
                    'Converting {} to {}.'.format(legacy_podcache_file_name, podcache_file_name))
                # Do not use the utils.parse_yaml as that has extra constructors
                # that should not be run when the cache file is being parsed.
                temp_data = yaml.load(
                    self.read_file(legacy_podcache_file_name)) or {}
                if 'objects' in temp_data and temp_data['objects']:
                    object_cache_file_name = '/{}'.format(
                        podcache.FILE_OBJECT_CACHE)
                    self.write_file(object_cache_file_name,
                                    json.dumps(temp_data['objects']))
                self.delete_file(legacy_podcache_file_name)

            if not self.file_exists(podcache_file_name):
                return {}
            try:
                return self.read_json(podcache_file_name) or {}
            except IOError:
                path = self.abs_path(podcache_file_name)
                raise podcache.PodCacheParseError(
                    'Error parsing: {}'.format(path))

    def _parse_routes_cache_file(self):
        with self.profile.timer('Pod._parse_routes_cache_file'):
            routes_cache_file_name = '{}{}'.format(
                self.PATH_CONTROL, podcache.FILE_ROUTES_CACHE)
            if not self.file_exists(routes_cache_file_name):
                return {}
            try:
                return self.read_json(routes_cache_file_name) or {}
            except ValueError:
                # File became corrupted; delete it and start over.
                # https://github.com/grow/grow/issues/1050#issuecomment-596346032
                self.delete_file(routes_cache_file_name)
                return {}
            except IOError:
                path = self.abs_path(routes_cache_file_name)
                raise podcache.PodCacheParseError(
                    'Error parsing: {}'.format(path))

    @utils.memoize
    def _parse_yaml(self):
        podspec_file_name = '/{}'.format(self.FILE_PODSPEC)
        try:
            return utils.parse_yaml(self.read_file(podspec_file_name))
        except IOError as e:
            path = self.abs_path(podspec_file_name)
            if e.args[0] == 2 and e.filename:
                raise PodDoesNotExistError('Pod not found in: {}'.format(path))
            raise podspec.PodSpecParseError('Error parsing: {}'.format(path))

    @staticmethod
    def clean_pod_path(pod_path):
        """Cleanup the pod path."""
        if not pod_path.startswith('/'):
            pod_path = '/{}'.format(pod_path)
        return pod_path

    def set_env(self, env):
        if not isinstance(env, environment.Env):
            env = environment.Env(env)
        if env and env.name:
            content = untag.Untag.untag(self._parse_yaml(), params={
                'env': untag.UntagParamRegex(env.name),
            })
            self._yaml = content
            # Preprocessors may depend on env, reset cache.
            # pylint: disable=no-member
            self.list_preprocessors.reset()
            self.podcache.reset()
            # Need to reload the extension configs for changes based on env.
            self._extensions_controller.update_extension_configs(
                self.yaml.get('ext', []))
        self.env = env

    @utils.cached_property
    def cache(self):
        return cachelib.SimpleCache(default_timeout=0)

    @property
    def error_routes(self):
        return self.yaml.get('error_routes')

    @property
    def exists(self):
        return self.file_exists('/{}'.format(self.FILE_PODSPEC))

    @property
    def experiments(self):
        return self._experiments

    @property
    def extensions_controller(self):
        return self._extensions_controller

    @property
    def grow_version(self):
        return self.podspec.grow_version

    @utils.cached_property
    def jinja_bytecode_cache(self):
        return jinja2.MemcachedBytecodeCache(client=self.cache)

    @property
    def logger(self):
        return logger.LOGGER

    @property
    def markdown(self):
        """Markdown processor with pod flavored markdown."""
        # Do not cached property, causes issue with extensions.
        return self.markdown_util.markdown

    @utils.cached_property
    def markdown_util(self):
        """Markdown util specific to pod flavored markdown."""
        return markdown_utils.MarkdownUtil(self)

    @utils.cached_property
    def partials(self):
        """Returns the pod partials object."""
        return partials.Partials(self)

    @utils.cached_property
    def path_filter(self):
        """Filter for testing path formats."""
        if self.env.name == environment.Name.DEV or self.env.name is None:
            filter_config = self.yaml.get('filter', {})
        else:
            if 'deployments' not in self.yaml:
                raise ValueError('No pod-specific deployments configured.')
            filter_config = (self.yaml['deployments']
                             .get(self.env.name, {})
                             .get('filter', {}))
        return grow_path_filter.PathFilter(
            filter_config.get('ignore_paths'), filter_config.get('include_paths'))

    @utils.cached_property
    def path_format(self):
        """Format utility for url paths."""
        return grow_path_format.PathFormat(self)

    @property
    def podcache(self):
        if not self._podcache:
            self._podcache = podcache.PodCache(
                dep_cache=self._parse_dep_cache_file(),
                obj_cache=self._parse_object_cache_file(),
                routes_cache=self._parse_routes_cache_file(),
                pod=self)
        return self._podcache

    @property
    def podspec(self):
        return podspec.PodSpec(yaml=self.yaml, pod=self)

    @utils.cached_property
    def profile(self):
        """Profile object for code timing."""
        return profile.Profile()

    @utils.cached_property
    def render_pool(self):
        """Render pool rendering documents."""
        return grow_render_pool.RenderPool(self)

    @utils.cached_property
    def router(self):
        """Router object for routing."""
        return grow_router.Router(self)

    @property
    def static_configs(self):
        """Yields each of the static configurations."""
        podspec_config = self.podspec.get_config()
        if 'static_dirs' not in podspec_config:
            return
        for config in podspec_config['static_dirs']:
            yield config

        # Allow for extensions to create static directory configs.
        hook_static_dirs = self.extensions_controller.trigger('static_dir')
        for config in hook_static_dirs:
            yield config

    @utils.cached_property
    def tmp_dir(self):
        """Temp directory for temporary file caching."""
        dir_name = tempfile.mkdtemp()
        _POD_TEMP_DIRS.append(dir_name)
        return dir_name

    @property
    def title(self):
        return self.yaml.get('title')

    @utils.cached_property
    def translation_stats(self):
        return translation_stats.TranslationStats()

    @property
    def extensions_dir(self):
        return self.yaml.get('extensions_dir', Pod.INSTALLED_EXTENSIONS_DIR_PATH)

    @property
    def ui(self):
        if self.env.name == environment.Name.DEV or self.env.name is None:
            ui_config = self.yaml.get('ui')
        else:
            if 'deployments' not in self.yaml:
                raise ValueError('No pod-specific deployments configured.')
            ui_config = (self.yaml['deployments']
                         .get(self.env.name, {})
                         .get('ui'))
        return ui_config

    @property
    def yaml(self):
        if self._yaml is utils.SENTINEL:
            self._yaml = self._parse_yaml() or {}
        return self._yaml

    def abs_path(self, pod_path):
        path = os.path.join(self.root, pod_path.lstrip('/'))
        return os.path.join(self.root, path)

    def copy_file_to(self, source_pod_path, destination_pod_path):
        source_path = self._normalize_path(source_pod_path)
        dest_path = self._normalize_path(destination_pod_path)
        return self.storage.copy_to(source_path, dest_path)

    def create_collection(self, collection_path, fields):
        pod_path = os.path.join(
            collection.Collection.CONTENT_PATH, collection_path)
        return collection.Collection.create(pod_path, fields, pod=self)

    def delete(self):
        """Deletes the pod by deleting all of its files."""
        pod_paths = self.list_dir('/')
        for path in pod_paths:
            self.delete_file(path)
        return pod_paths

    def delete_file(self, pod_path):
        path = self._normalize_path(pod_path)
        return self.storage.delete(path)

    def delete_files(self, pod_paths, recursive=False, pattern=None):
        """Delete matching files from the pod_paths."""
        normal_paths = []
        for pod_path in pod_paths:
            normal_paths.append(self._normalize_path(pod_path))
        return self.storage.delete_files(normal_paths, recursive=recursive, pattern=pattern)

    def disable(self, feature):
        """Disable a grow feature."""
        self._features.disable(feature)

    def dump(self, suffix='index.html', append_slashes=True, pod_paths=None,
             use_threading=True, source_dir=None):
        """Dumps the pod, yielding rendered_doc based on pod routes."""
        for rendered_doc in self.export(
                suffix=suffix, append_slashes=append_slashes, pod_paths=pod_paths,
                use_threading=use_threading, source_dir=source_dir):
            yield rendered_doc
        if self.ui and self.is_enabled(self.FEATURE_UI):
            for rendered_doc in self.export_ui():
                yield rendered_doc

    def enable(self, feature):
        """Enable a grow feature."""
        self._features.enable(feature)

    def export(self, suffix=None, append_slashes=False, pod_paths=None,
               use_threading=True, source_dir=None):
        """Builds the pod, yielding rendered_doc based on pod routes."""
        for rendered_doc in renderer.Renderer.rendered_docs(
                self, self.router.routes, use_threading=use_threading,
                source_dir=source_dir):
            yield rendered_doc

    def export_ui(self):
        """Builds the grow ui tools, returning a mapping of paths to content."""
        paths = []
        source_prefix = 'node_modules/'
        destination_root = '_grow/ui/'
        tools_dir = 'tools/'
        tool_prefix = 'grow-tool-'

        # Add the base ui files.
        source_root = os.path.join(utils.get_grow_dir(), 'ui', 'dist')
        for path in ['css/ui.min.css', 'js/ui.min.js']:
            source_path = os.path.join(source_root, path)
            output_path = os.sep + os.path.join(destination_root, path)
            yield rendered_document.RenderedDocument(
                output_path, self.storage.read(source_path))

        # Add the files from each of the tools.
        for tool in self.ui.get('tools', []):
            tool_path = '{}{}{}'.format(
                source_prefix, tool_prefix, tool['kind'])
            for root, dirs, files in self.walk(tool_path):
                for directory in dirs:
                    if directory.startswith('.'):
                        dirs.remove(directory)
                pod_dir = root.replace(self.root, '')
                for file_name in files:
                    paths.append(os.path.join(pod_dir, file_name))

        text = 'Building UI Tools: %(value)d/{} (in %(time_elapsed).9s)'
        widgets = [progressbar.FormatLabel(text.format(len(paths)))]
        progress = progressbar_non.create_progressbar(
            "Building UI Tools...", widgets=widgets, max_value=len(paths))
        progress.start()
        for path in paths:
            output_path = path.replace(
                source_prefix, '{}{}'.format(destination_root, tools_dir))
            yield rendered_document.RenderedDocument(
                output_path, self.read_file(path))
            progress.update(progress.value + 1)
        progress.finish()

    def file_exists(self, pod_path):
        path = self._normalize_path(pod_path)
        return self.storage.exists(path)

    def file_modified(self, pod_path):
        path = self._normalize_path(pod_path)
        return self.storage.modified(path)

    def file_size(self, pod_path):
        path = self._normalize_path(pod_path)
        return self.storage.size(path)

    def get_catalogs(self, template_path=None):
        return catalog_holder.Catalogs(pod=self, template_path=template_path)

    def get_collection(self, collection_path):
        """Returns a collection.

        Args:
          collection_path: A collection's path relative to the /content/ directory.
        Returns:
          Collection.
        """
        pod_path = os.path.join(
            collection.Collection.CONTENT_PATH, collection_path)
        cached = self.podcache.collection_cache.get_collection(pod_path)
        if cached:
            return cached

        col = collection.Collection.get(pod_path, _pod=self)
        self.podcache.collection_cache.add_collection(col)
        return col

    def get_deployment(self, nickname):
        """Returns a pod-specific deployment."""
        # Lazy import avoids environment errors and speeds up importing.
        from grow.deployments import deployments as grow_deployments
        if 'deployments' not in self.yaml:
            raise ValueError('No pod-specific deployments configured.')
        destination_configs = self.yaml['deployments']
        if nickname not in destination_configs:
            text = 'No deployment named {}. Valid deployments: {}.'
            keys = ', '.join(list(destination_configs.keys()))
            raise ValueError(text.format(nickname, keys))
        deployment_params = destination_configs[nickname]
        kind = deployment_params.pop('destination')
        try:
            config = destination_configs[nickname]
            deployments = grow_deployments.Deployments()
            self.extensions_controller.trigger(
                'deployment_register', deployments)
            deployment = deployments.make_deployment(
                kind, config, name=nickname)
        except TypeError:
            logging.exception('Invalid deployment parameters.')
            raise
        deployment.pod = self
        return deployment

    def get_doc(self, pod_path, locale=None):
        """Returns a document, given the document's pod path."""
        collection_path, unused_path = os.path.split(pod_path)
        if not collection_path or not unused_path:
            text = '"{}" is not a path to a document.'.format(pod_path)
            raise collection.BadCollectionNameError(text)
        original_collection_path = collection_path
        col = self.get_collection(collection_path)
        while not col.exists:
            col = col.parent
            if not col:
                col = self.get_collection(original_collection_path)
                break
        doc = col.get_doc(pod_path, locale=locale)
        return doc

    def get_home_doc(self):
        home = self.yaml.get('home')
        if home is None:
            return None
        return self.get_doc(home)

    @utils.memoize
    def get_jinja_env(self, locale='', root=None):
        with self._jinja_env_lock:
            kwargs = {
                'autoescape': True,
                'extensions': [
                    'jinja2.ext.autoescape',
                    'jinja2.ext.do',
                    'jinja2.ext.i18n',
                    'jinja2.ext.loopcontrols',
                    'jinja2.ext.with_',
                ],
                'loader': self.storage.JinjaLoader(self.root if root is None else root),
                'lstrip_blocks': True,
                'trim_blocks': True,
            }
            if self.env.cached:
                kwargs['bytecode_cache'] = self._get_bytecode_cache()
            kwargs['extensions'].extend(self.list_jinja_extensions())
            env = jinja_dependency.DepEnvironment(**kwargs)
            env.filters.update(filters.create_builtin_filters(env, self, locale=locale))
            env.globals.update(
                **tags.create_builtin_globals(env, self, locale=locale))
            return env

    def get_static(self, pod_path, locale=None):
        """Returns a StaticFile, given the static file's pod path."""
        document = static_document.StaticDocument(
            self, pod_path, locale=locale)
        if document.exists:
            return document

        text = ('Either no file exists at "{}" or the "static_dirs" setting was '
                'not configured for this path in {}.'.format(
                    pod_path, self.FILE_PODSPEC))
        raise errors.DocumentDoesNotExistError(text)

    def get_podspec(self):
        return self.podspec

    @utils.memoize
    def get_translator(self, service=utils.SENTINEL):
        if 'translators' not in self.yaml:
            return None
        if ('services' not in self.yaml['translators']
                or not self.yaml['translators']['services']):
            return None
        translator_config = self.yaml['translators']
        translator_extensions = self.yaml.get(
            'extensions', {}).get('translators', [])
        translators.register_extensions(
            translator_extensions,
            self.root,
        )

        # TODO: Show deprecation message in the future.
        # if translator_extensions:
        #     legacy_message = 'Legacy translators are deprecated and will be removed in the future: {}'
        #     self.deprecated(
        #         'legacy_translator',
        #         legacy_message.format(', '.join(translator_extensions)),
        #         url='https://grow.dev/migration/1.0.0')

        translator_services = copy.deepcopy(translator_config['services'])
        if service is not utils.SENTINEL:
            valid_service_kinds = [each['service']
                                   for each in translator_services]
            if not valid_service_kinds:
                text = 'Missing required "service" field in translator config.'
                raise ValueError(text)
            if service not in valid_service_kinds and service is not None:
                text = 'No translator service "{}". Valid services: {}.'
                keys = ', '.join(valid_service_kinds)
                raise ValueError(text.format(service, keys))
        else:
            # Use the first configured translator by default.
            translator_services = translator_services[:1]
        for service_config in translator_services:
            if service_config.get('service') == service or len(translator_services) == 1:
                translator_kind = service_config.pop('service')
                return translators.create_translator(
                    self, translator_kind, service_config,
                    project_title=translator_config.get('project_title'),
                    instructions=translator_config.get('instructions'))
        raise ValueError('No translator service found: {}'.format(service))

    def get_url(self, pod_path, locale=None):
        if pod_path.startswith('/content'):
            doc = self.get_doc(pod_path, locale=locale)
        else:
            doc = self.get_static(pod_path, locale=locale)

        if not doc.exists:
            raise errors.DocumentDoesNotExistError(
                'Referenced document does not exist: {}'.format(pod_path))
        return doc.url

    def hash_file(self, pod_path):
        """Provide the hash of the file from the storage."""
        path = self._normalize_path(pod_path)
        with self.profile.timer('Pod.hash_file', label=path, meta={'path': path}):
            return self.storage.hash(path)

    def inject_preprocessors(self, doc=None, collection=None):
        """Conditionally injects or creates data from preprocessors. If a doc
        is provided, preprocessors inject data to doc fields. If a collection
        is provided, the first matching preprocessor returns a list of docs."""
        for preprocessor in self.list_preprocessors():
            if doc is not None:
                if preprocessor.can_inject(doc=doc):
                    preprocessor.inject(doc=doc)
                    return preprocessor
            if collection is not None:
                if preprocessor.can_inject(collection=collection):
                    preprocessor.inject(collection=collection)
                    return preprocessor

    def is_enabled(self, feature):
        return self._features.is_enabled(feature)

    def list_collections(self, paths=None):
        cols = collection.Collection.list(self)
        if paths:
            return [col for col in cols if col.collection_path in paths]
        return cols

    def list_deployments(self):
        destination_configs = self.yaml['deployments']
        results = []
        for name in destination_configs.keys():
            results.append(self.get_deployment(name))
        return results

    def list_dir(self, pod_path='/', recursive=True):
        path = self._normalize_path(pod_path)
        return self.storage.listdir(path, recursive=recursive)

    def list_jinja_extensions(self):
        loaded_extensions = []
        jinja_extensions = self.yaml.get('extensions', {}).get('jinja2', [])
        for name in jinja_extensions:
            try:
                value = extension_importer.ExtensionImporter.find_extension(
                    name, self.root)
            except ImportError:
                logging.error(
                    'Error importing %s. Module path must be relative to '
                    'the pod root.', repr(name))
                raise
            loaded_extensions.append(value)

        # TODO: Show deprecation message in the future.
        # if jinja_extensions:
        #     legacy_message = 'Legacy jinja2 extensions are deprecated and will be removed in the future: {}'
        #     self.deprecated(
        #         'legacy_jinja2',
        #         legacy_message.format(', '.join(jinja_extensions)),
        #         url='https://grow.dev/migration/1.0.0')

        loaded_extensions.extend(
            self.extensions_controller.trigger('jinja_extensions'))
        return loaded_extensions

    def list_locales(self):
        codes = self.yaml.get('localization', {}).get('locales', [])
        return self.normalize_locales(locales.Locale.parse_codes(codes))

    @utils.memoize
    def list_preprocessors(self):
        results = []
        preprocessors.register_extensions(
            self.yaml.get('extensions', {}).get('preprocessors', []),
            self.root,
        )
        preprocessor_config = copy.deepcopy(self.yaml.get('preprocessors', []))
        legacy_preprocessors = []
        for params in preprocessor_config:
            kind = params.pop('kind')
            legacy_preprocessors.append(kind)

            try:
                preprocessor = preprocessors.make_preprocessor(
                    kind, params, self)
                results.append(preprocessor)
            except ValueError as err:
                # New extensions don't exists and are considered a value error.
                pass

        # TODO: Show deprecation message in the future.
        # if legacy_preprocessors:
        #     legacy_message = 'Legacy preprocessors are deprecated and will be removed in the future: {}'
        #     self.deprecated(
        #         'legacy_preprocessors',
        #         legacy_message.format(', '.join(legacy_preprocessors)),
        #         url='https://grow.dev/migration/1.0.0')

        return results

    def list_statics(self, pod_path, locale=None, include_hidden=False):
        for path in self.list_dir(pod_path):
            if include_hidden or not path.rsplit(os.sep, 1)[-1].startswith('.'):
                yield self.get_static(pod_path + path, locale=locale)

    def match(self, path):
        return self.router.routes.match(path)

    def move_file_to(self, source_pod_path, destination_pod_path):
        source_path = self._normalize_path(source_pod_path)
        dest_path = self._normalize_path(destination_pod_path)
        return self.storage.move_to(source_path, dest_path)

    def normalize_locale(self, locale, default=None):
        locale = locale or default or self.podspec.default_locale
        if isinstance(locale, str):
            try:
                locale = locales.Locale.parse(locale)
            except ValueError as err:
                # Locale could be an alias.
                locale = locales.Locale.from_alias(self, locale)
        if locale is not None:
            locale.set_alias(self)
        return locale

    def normalize_locales(self, locale_list):
        for locale in locale_list:
            locale.set_alias(self)
        return locale_list

    def open_file(self, pod_path, mode=None):
        path = self._normalize_path(pod_path)
        return self.storage.open(path, mode=mode)

    def preprocess(self, preprocessor_names=None, run_all=False, tags=None,
                   build=True, ratelimit=None):
        if not preprocessor_names:
            self.catalogs.compile()  # Preprocess translations.

        # Extension support for preprocessors.
        preprocessors = self.yaml.get('preprocessors', [])
        for config in preprocessors:
            self.extensions_controller.trigger(
                'preprocess', config, preprocessor_names, tags, run_all, ratelimit)

        # Legacy support for preprocessors.
        # TODO Remove when not supporting the legacy preprocessors.
        for preprocessor in self.list_preprocessors():
            if preprocessor_names:
                if preprocessor.name in preprocessor_names:
                    preprocessor.run(build=build)
            elif tags:
                if set(preprocessor.tags).intersection(tags):
                    preprocessor.run(build=build)
            elif preprocessor.autorun or run_all:
                preprocessor.run(build=build)
            if ratelimit:
                time.sleep(ratelimit)

    def read_csv(self, path, locale=utils.SENTINEL):
        with self.profile.timer('Pod.read_csv', label=path, meta={'path': path}):
            return utils.get_rows_from_csv(pod=self, path=path, locale=locale)

    def read_file(self, pod_path):
        path = self._normalize_path(pod_path)
        with self.profile.timer('Pod.read_file', label=path, meta={'path': path}):
            return self.storage.read(path)

    def read_json(self, path):
        """Read and parse a json file."""
        with self.open_file(path, 'r') as json_file:
            return json.load(json_file)

    def read_yaml(self, path, locale=None):
        """Read, parse, and untag a yaml file."""
        contents = self.podcache.file_cache.get(path, locale=locale)
        if contents is None:
            label = '{} ({})'.format(path, locale)
            meta = {'path': path, 'locale': locale}
            with self.profile.timer('Pod.read_yaml', label=label, meta=meta):
                fields = self.podcache.file_cache.get(path, locale='__raw__')
                if fields is None:
                    fields = utils.parse_yaml(
                        self.read_file(path), pod=self, locale=locale)
                    self.podcache.file_cache.add(
                        path, fields, locale='__raw__')
                try:
                    contents = self.untag(fields, locale=locale)
                    self.podcache.file_cache.add(path, contents, locale=locale)
                except Exception:
                    logging.error('Error parsing -> {}'.format(path))
                    raise
        return contents

    def reset_yaml(self):
        # Mark that the yaml needs to be reparsed.
        self._yaml = utils.SENTINEL
        # Tell the cached property to reset.
        # pylint: disable=no-member
        self._parse_yaml.reset()

    def untag(self, contents, locale=None):
        """Untag data using the pod specific untagging params."""
        return untag.Untag.untag(
            contents, locale_identifier=locale, params={
                'env': untag.UntagParamRegex(self.env.name),
            })

    def walk(self, pod_path):
        path = self._normalize_path(pod_path)
        return self.storage.walk(path)

    def write_file(self, pod_path, content):
        with self.profile.timer(
                'Pod.write_file', label=pod_path, meta={'path': pod_path}):
            self.podcache.file_cache.remove(pod_path)
            path = self._normalize_path(pod_path)
            self.storage.write(path, content)

    def write_yaml(self, path, content):
        with self.profile.timer(
                'Pod.write_yaml', label=path, meta={'path': path}):
            self.podcache.collection_cache.remove_by_path(path)
            self.podcache.document_cache.remove_by_path(path)
            content = utils.dump_yaml(content)
            self.write_file(path, content)



"""Deprecation helper for warning users of deprecated classes."""

import logging

# pylint: disable=too-few-public-methods
class DeprecationHelper(object):
    """Deprecation helper class for deprecating a class."""

    def __init__(self, new_target, message, warn=logging.warn):
        self.new_target = new_target
        self.message = message
        self._warning = warn
        self._has_warned = False

    def _warn(self):
        # Only warn once to prevent spamming logs.
        if not self._has_warned:
            self._warning(self.message)
            self._has_warned = True

    def __call__(self, *args, **kwargs):
        self._warn()
        return self.new_target(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self.new_target, attr)

# pylint: disable=too-few-public-methods
class DeprecationManager(object):
    """Deprecation manager class for deprecation messages without repeating."""

    def __init__(self, warn=logging.warn):
        self._warning = warn
        self._displayed = set()
        self.extra_url_message = '   Visit {} for more information.'

    def __call__(self, *args, **kwargs):
        self.warn(*args, **kwargs)

    def warn(self, key, message, url=None):
        # Only warn once to prevent spamming logs.
        if key in self._displayed:
            return

        if url:
            url_message = self.extra_url_message.format(url)
            message = '{}\n{}'.format(message, url_message)

        self._displayed.add(key)
        self._warning(message)


# pylint: disable=too-few-public-methods
class MovedHelper(DeprecationHelper):
    """Specialized deprecation helper for warning of moved classes."""

    def __init__(self, new_target, orig_class_name, *args, **kwargs):
        new_class_name = new_target.__module__ + "." + new_target.__name__
        message = 'The {} class has moved to {} and will be removed in a future version.'.format(
            orig_class_name, new_class_name)
        super(MovedHelper, self).__init__(new_target, message, *args, **kwargs)



"""Dev manager message hook."""

from grow.extensions.hooks import base_hook


class DevManagerMessageHook(base_hook.BaseHook):
    """Hook for dev server messaging."""

    KEY = 'dev_manager_message'
    NAME = 'Dev Manager Message Handler'

    # display_func(label, message, label_color=None, message_color=None)
    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, display_func, url_base, url_root, *_args, **_kwargs):
        """Trigger the dev handler hook."""
        return previous_result



"""Markdown utility methods for pods."""

import copy
import markdown
from markdown.extensions import tables
from grow.common import markdown_extensions
from grow.common import structures
from grow.common import utils


class MarkdownUtil(object):
    """Utility class for working with pod flavored markdown."""

    def __init__(self, pod):
        self.pod = pod

    @property
    def extensions(self):
        """List of enabled extensions for the pod."""
        # Do not cache property so that the extensions are created fresh.
        extensions = [
            tables.TableExtension(),
            markdown_extensions.TocExtension(pod=self.pod),
            markdown_extensions.CodeBlockExtension(self.pod),
            markdown_extensions.IncludeExtension(self.pod),
            markdown_extensions.UrlExtension(self.pod),
            'markdown.extensions.fenced_code',
            'markdown.extensions.codehilite',
        ]

        for config in self.markdown_config:
            if config['kind'] in extensions:
                continue

            if config['kind'].startswith('markdown.extensions'):
                extensions.append(config['kind'])

        return extensions

    @utils.cached_property
    def extension_configs(self):
        """Extension configurations from the podspec."""
        extension_configs = {}

        for config in self.markdown_config:
            if config['kind'].startswith('markdown.extensions'):
                ext_config = copy.deepcopy(config)
                ext_config.pop('kind', None)
                if ext_config:
                    extension_configs[config['kind']] = ext_config

        # Special handling for code highlighting backwards compatability.
        config = self.extension_config('markdown.extensions.codehilite')
        codehilite_config = {
            'pygments_style': 'default',
            'noclasses': True,
            'css_class': 'code',
        }
        if 'theme' in config:
            codehilite_config['pygments_style'] = config.theme
        if 'classes' in config:
            codehilite_config['noclasses'] = not config.classes
        if 'class_name' in config:
            codehilite_config['css_class'] = config.class_name
        extension_configs['markdown.extensions.codehilite'] = codehilite_config

        return extension_configs

    @property
    def markdown(self):
        """Markdown object using the pod configuration."""
        return markdown.Markdown(
            extensions=self.extensions,
            extension_configs=self.extension_configs)

    @utils.cached_property
    def markdown_config(self):
        """Markdown config from podspec."""
        if 'markdown' in self.pod.podspec:
            markdown_config = self.pod.podspec.markdown
            if 'extensions' in markdown_config:
                return markdown_config['extensions']
        return []

    def extension_config(self, kind):
        """Get the markdown config for a specific extension."""
        for extension in self.markdown_config:
            if extension.get('kind', '') == kind:
                return structures.AttributeDict(extension)
        return structures.AttributeDict({})



"""Routing information for the collection."""

import os


class CollectionRoutes(object):
    """Collection level routing information."""

    ROUTES_PATH = '_routes.yaml'

    def __getattr__(self, attr):
        def _get_field(pod_path, default_value=None):
            return self.get_field(pod_path, attr, default_value)
        return _get_field

    def __init__(self, pod_paths=None):
        self.pod_paths = pod_paths or {}

    def _get_meta(self, pod_path):
        pod_path = pod_path.lstrip(os.path.sep)
        return self.pod_paths.get(pod_path, {})

    def get_field(self, pod_path, field, default_value=None):
        """Get document field from metadata."""
        data = self._get_meta(pod_path)
        return data.get(field, default_value)



"""Base Config control class."""


class BaseConfig(object):
    """Base class for identifier based configuration management."""

    def __init__(self, config=None):
        self._config = config or {}

    @staticmethod
    def _split_identifier(identifier):
        return identifier.split('.')

    def export(self):
        """Export the raw config."""
        return self._config

    def get(self, identifier, default_value=None):
        """Retrieve the identifier value in the config."""
        parts = self._split_identifier(identifier)
        item = self._config
        for part in parts:
            if part not in item:
                return default_value
            item = item[part]
        return item

    def prefixed(self, prefix):
        """Create a utility for accessing config with a common identifier prefix."""
        return BaseConfigPrefixed(self, prefix)

    def set(self, identifier, value):
        """Set the value of the identifier in the config."""
        parts = self._split_identifier(identifier)
        key = parts.pop()
        item = self._config
        for part in parts:
            if part not in item:
                item[part] = {}
            item = item[part]
        item[key] = value


class BaseConfigPrefixed(object):
    """Utility class for shortcutting common prefixes in identifiers."""

    def __init__(self, config, prefix):
        self._config = config
        self.prefix = self.normalize_prefix(prefix)

    @staticmethod
    def normalize_prefix(prefix):
        """Normalize how the prefix is formatted."""
        prefix = prefix.strip()
        if not prefix:
            return ''
        if not prefix.endswith('.'):
            prefix = '{}.'.format(prefix)
        return prefix

    def prefix_identifier(self, identifier):
        """Adds the prefix to the identifier"""
        return '{}{}'.format(self.prefix, identifier)

    def get(self, identifier, default_value=None):
        """Retrieve the identifier value in the config with prefix."""
        return self._config.get(self.prefix_identifier(identifier), default_value=default_value)

    def set(self, identifier, value):
        """Set the value of the identifier in the config with prefix."""
        return self._config.set(self.prefix_identifier(identifier), value=value)



"""Base class for the deployment destination registration hook."""

from grow.extensions.hooks import base_hook


class DeploymentRegisterHook(base_hook.BaseHook):
    """Hook for deployment destination registration."""

    KEY = 'deployment_register'
    NAME = 'Deployment Destination Registration'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, deployments, *_args, **_kwargs):
        """Trigger the deployment destination registration hook."""
        if previous_result:
            return previous_result
        # Return None if nothing has changed.
        return None



"""Base class for the router add hook."""

from grow.extensions.hooks import base_hook


class RouterAddHook(base_hook.BaseHook):
    """Hook for router add."""

    KEY = 'router_add'
    NAME = 'Router Add'

    # pylint: disable=arguments-differ,unused-argument
    def trigger(self, previous_result, router, *_args, **_kwargs):
        """Trigger the router add hook."""
        if previous_result:
            return previous_result
        return None



# -*- coding: utf-8 -*-
"""
Utility for converting Grow sites to enable collection level routing. Instead
of storing things that affect routing directly in the document it is stored
at the level of the collection to prevent the need to read every document to
generate the routing information.

WARNING: This feature is used in the `separate_routing` experiment.

Usage:

    grow convert --type=collection_routing
"""

import collections
import logging
import os
import yaml
from grow.common import yaml_utils


ROUTES_FILENAME = '_blueprint.yaml'
COLLECTION_META_KEYS = ('$path', '$localization', '$view')
COLLECTION_BLUEPRINT_KEYS = (
    '$path', '$localization', '$view', 'path', 'localization', 'view',
    'composite_files')


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class RoutesData(object):
    """Store and format the routes information pulled from the documents."""

    def __init__(self, collection_path, blueprint):
        self.paths = collections.OrderedDict()
        self.collection_path = collection_path
        self.blueprint = blueprint

    @property
    def data(self):
        """The yaml contents to write to the file."""
        data = collections.OrderedDict()

        # Pull over the existing blueprint data.
        if '$path' in self.blueprint or 'path' in self.blueprint:
            data['path'] = self.blueprint.get(
                '$path', self.blueprint.get('path'))

        tagged_keys = tuple(['{}@'.format(key)
                             for key in COLLECTION_BLUEPRINT_KEYS])

        for key in sorted(self.blueprint.keys()):
            if key in COLLECTION_BLUEPRINT_KEYS or key.startswith(tagged_keys):
                data[key.lstrip('$')] = self.blueprint[key]

        data['routes'] = self.paths
        return data

    def extract_doc(self, doc):
        """Extract the meta information from the doc to use for the routes."""
        if doc.pod_path.endswith('.yaml'):
            raw_data = yaml.load(
                doc.pod.read_file(doc.pod_path), Loader=yaml_utils.PlainTextYamlLoader)
        else:
            raw_data = doc.format.front_matter.raw_data

        if not raw_data:
            print('No raw data found for document: {}'.format(doc.pod_path))
            return

        data = collections.OrderedDict()

        tagged_keys = tuple(['{}@'.format(key)
                             for key in COLLECTION_META_KEYS])

        for key, value in raw_data.items():
            if key in COLLECTION_META_KEYS or key.startswith(tagged_keys):
                normal_key = key.lstrip('$')
                if 'path' in key:
                    # Use `__NONE` as sentinel to support equating a `value` of None.
                    collection_path = self.blueprint.get(
                            key, self.blueprint.get(normal_key, '__NONE'))
                    # Skip the paths that are the same as the collection.
                    if collection_path == value:
                        continue
                data[normal_key] = value

        if data:
            self.paths[doc.collection_sub_path[1:]] = data

    def write_routes(self, pod, collection):
        """Write the converted routes to the configuration file."""
        routes_file = os.path.join(collection.pod_path, ROUTES_FILENAME)

        if self.data['routes']:
            print(' └─ Writing: {}'.format(routes_file))
            print('')
            output = yaml.dump(
                self.data, Dumper=yaml_utils.PlainTextYamlDumper,
                default_flow_style=False, allow_unicode=True, width=800)
            pod.write_file(routes_file, output)
        else:
            print(' └─ Skipping: {}'.format(routes_file))
            print('')


class ConversionCollection(object):
    """Temporary collection class for doing a conversion for a collection."""

    def __init__(self, pod, collection):
        self.pod = pod
        self.collection = collection
        self.routes_data = RoutesData(collection.pod_path, collection.yaml)

    def convert(self):
        """Perform the conversion to use collection based routing."""
        print('Converting: {}'.format(self.collection.pod_path))

        # Pull out the meta information from all the docs.
        sorted_docs = sorted(self.collection.list_docs_unread(), key=lambda doc: doc.pod_path)
        for doc in sorted_docs:
            self.routes_data.extract_doc(doc)

        self.routes_data.write_routes(self.pod, self.collection)


class Converter(object):

    @staticmethod
    def convert(pod):
        logging.info('Converting pod for collection level routing: {}'.format(
            pod.root))

        for collection in pod.list_collections():
            conv_collection = ConversionCollection(pod, collection)
            conv_collection.convert()



"""Track the jinja2 rendering dependencies using a custom environment.

Trying to track with a custom loader does not work since the templates are cached
by the environment and only triggers the loader the first time the template is used.

By using a modified version of the code generator and environment, the context
is passed through to `_load_template` method which normally does not need the context.
Since the `_load_template` method is called before the template caching and does
not affect the performance of the template cache.

When updating jinja to a new version:

- The `visit_Include` adds `, context` as an argument to the function.
- The `get_template`, `select_template`, `get_or_select_template` pass the `context`
  argument to the `_load_template` method.
- The `_load_template` method can remain unchanged.
"""

from jinja2 import nodes
from jinja2.compiler import supports_yield_from, CodeGenerator
from jinja2.environment import Environment, Template, TemplateNotFound, TemplatesNotFound
from jinja2._compat import string_types


class DepCodeGenerator(CodeGenerator):
    """Custom code generator for dependency tracking."""

    def visit_Include(self, node, frame):
        """Copied from upstream. Added context arg to `func` call."""
        if node.ignore_missing:
            self.writeline('try:')
            self.indent()

        func_name = 'get_or_select_template'
        if isinstance(node.template, nodes.Const):
            if isinstance(node.template.value, string_types):
                func_name = 'get_template'
            elif isinstance(node.template.value, (tuple, list)):
                func_name = 'select_template'
        elif isinstance(node.template, (nodes.Tuple, nodes.List)):
            func_name = 'select_template'

        # print 'Visit Include'
        self.writeline('template = environment.%s(' % func_name, node)
        self.visit(node.template, frame)
        # Changed from upstream.
        self.write(', %r, context=context)' % self.name)
        if node.ignore_missing:
            self.outdent()
            self.writeline('except TemplateNotFound:')
            self.indent()
            self.writeline('pass')
            self.outdent()
            self.writeline('else:')
            self.indent()

        skip_event_yield = False
        if node.with_context:
            loop = self.environment.is_async and 'async for' or 'for'
            self.writeline('%s event in template.root_render_func('
                           'template.new_context(context.get_all(), True, '
                           '%s)):' % (loop, self.dump_local_context(frame)))
        elif self.environment.is_async:
            self.writeline('for event in (await '
                           'template._get_default_module_async())'
                           '._body_stream:')
        else:
            if supports_yield_from:
                self.writeline('yield from template._get_default_module()'
                               '._body_stream')
                skip_event_yield = True
            else:
                self.writeline('for event in template._get_default_module()'
                               '._body_stream:')

        if not skip_event_yield:
            self.indent()
            self.simple_write('event', frame)
            self.outdent()

        if node.ignore_missing:
            self.outdent()


class DepEnvironment(Environment):
    """Custom environment for dependency tracking."""

    code_generator_class = DepCodeGenerator

    def get_template(self, name, parent=None, globals=None, context=None):
        """Copied from upstream. Added context arg passthrough."""
        if isinstance(name, Template):
            return name
        if parent is not None:
            name = self.join_path(name, parent)
        return self._load_template(name, self.make_globals(globals), context=context)

    def select_template(self, names, parent=None, globals=None, context=None):
        """Copied from upstream. Added context arg passthrough."""
        if not names:
            raise TemplatesNotFound(message='Tried to select from an empty list '
                                            'of templates.')
        globals = self.make_globals(globals)
        for name in names:
            if isinstance(name, Template):
                return name
            if parent is not None:
                name = self.join_path(name, parent)
            try:
                return self._load_template(name, globals, context=context)
            except TemplateNotFound:
                pass
        raise TemplatesNotFound(names)

    def get_or_select_template(self, template_name_or_list,
                               parent=None, globals=None, context=None):
        """Copied from upstream. Added context arg passthrough."""
        if isinstance(template_name_or_list, string_types):
            return self.get_template(template_name_or_list, parent, globals, context=context)
        elif isinstance(template_name_or_list, Template):
            return template_name_or_list
        return self.select_template(template_name_or_list, parent, globals, context=context)

    # pylint: disable=redefined-builtin
    def _load_template(self, name, globals, context=None):
        """Custom template loading that tracks template as a dependency."""
        if context and '_track_dependency' in context:
            context['_track_dependency'](name)
        return super(DepEnvironment, self)._load_template(name, globals)



"""Deployment stats tracking."""

import collections
import os
import texttable
from protorpc import protojson
from . import messages


class Stats(object):

    def __init__(self, pod, paths=None, full=True):
        self.full = full
        self.pod = pod
        self.paths = paths

    def get_num_files_per_type(self):
        file_counts = collections.defaultdict(int)
        if self.paths:
            for path in self.paths:
                ext = os.path.splitext(path)[-1]
                file_counts[ext] += 1
        ms = []
        for ext, count in file_counts.items():
            ms.append(messages.FileCountMessage(ext=ext, count=count))
        return ms

    def to_message(self):
        message = messages.StatsMessage()
        collections = self.pod.list_collections()
        message.num_collections = len(collections)
        message.num_files_per_type = self.get_num_files_per_type()
        message.locales = [str(locale) for locale in self.pod.list_locales()]
        message.langs = self.pod.catalogs.list_locales()
        catalog = self.pod.catalogs.get_template()
        message.num_messages = len(catalog)
        return message

    def to_string(self):
        return protojson.encode_message(self.to_message())

    def to_tables(self):
        results = []

        table = texttable.Texttable(max_width=0)
        table.set_deco(texttable.Texttable.HEADER)
        rows = []
        rows.append(['Resource', 'Count'])

        all_collections = self.pod.list_collections()
        rows.append(['Collections', len(all_collections)])
        documents = []
        for collection in all_collections:
            documents += collection.list_docs()
        rows.append(['Documents', len(documents)])
        locales = self.pod.list_locales()
        rows.append(['Locales', len(locales)])
        self.pod.router.use_simple()
        self.pod.router.add_all()
        rows.append(['Routes', len(self.pod.router.routes)])
        template = self.pod.catalogs.get_template()
        rows.append(['Messages', len(template)])
        table.add_rows(rows)
        content = table.draw()
        results.append(content)

        if self.full:
            table = texttable.Texttable(max_width=0)
            table.set_deco(texttable.Texttable.HEADER)
            rows = []
            rows.append(['File type', 'Count'])
            exts_and_counts = self.get_num_files_per_type()
            exts_and_counts = sorted(exts_and_counts,
                                     key=lambda message: -message.count)
            for message in exts_and_counts:
                ext = message.ext or '.html'
                rows.append([ext, message.count])
            table.add_rows(rows)
            content = table.draw()
            results.append(content)

        table = texttable.Texttable(max_width=0)
        table.set_deco(texttable.Texttable.HEADER)
        catalogs = sorted(self.pod.catalogs, key=str)
        rows = []
        rows.append(['Translations ({})'.format(len(catalogs)), 'Messages'])
        for catalog in catalogs:
            num_messages = len(catalog)
            untranslated_messages = catalog.list_untranslated()
            translated_messages = num_messages - len(untranslated_messages)
            label = '{} / {}'.format(translated_messages, num_messages)
            rows.append([str(catalog.locale), label])

        table.add_rows(rows)
        if catalogs:
            content = table.draw()
            results.append(content)

        return results



import os


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class PreprocessorError(Error):
    pass


class BasePreprocessor(object):

    def __init__(self, pod, config, autorun=True, name=None, tags=[],
                 inject=False):
        self.pod = pod
        self.root = pod.root
        self.config = config
        self.logger = self.pod.logger
        self.autorun = autorun
        self.name = name
        self.tags = tags or []
        self.injected = inject

    def first_run(self):
        self.run()

    def run(self, build=True):
        raise NotImplementedError

    def list_watched_dirs(self):
        return []

    def normalize_path(self, path):
        if path.startswith('/'):
            return os.path.join(self.root, path.lstrip('/'))
        return path

    def can_inject(self, doc=None, collection=None):
        """Returns whether the preprocessor can inject data into a doc."""
        return False

    def inject(self, doc=None, collection=None):
        """Injects data into a doc without modifying the filesystem."""

    def get_edit_url(self, doc=None):
        """Returns the URL to edit a document in an external editor. Only
        applicable for preprocessors that inject data."""



"""Utility functions for managing the sdk."""

import os
import platform


PLATFORM = None
if 'Linux' in platform.system():
    PLATFORM = 'linux'
elif 'Darwin' in platform.system():
    PLATFORM = 'mac'
elif 'Windows' in platform.system():
    PLATFORM = 'win'


def subprocess_args(pod, shell=False):
    """Arguments for running subprocess commands."""
    env = os.environ.copy()
    node_modules_path = os.path.join(pod.root, 'node_modules', '.bin')
    env['PATH'] = str(env['PATH'] + os.path.pathsep + node_modules_path)
    if pod.env and pod.env.name:
        env['GROW_ENVIRONMENT_NAME'] = pod.env.name
    args = {
        'cwd': pod.root,
        'env': env,
    }
    if shell or PLATFORM == 'win':
        args['shell'] = True
    return args



"""Click deprecated groups."""

import logging
import click
from click.utils import make_str


# pylint: disable=too-few-public-methods
class DeprecatedItem(object):
    """Configuration for an alias item."""

    def __init__(self, old_name, new_name, cmd=None):
        self.old_name = old_name
        self.new_name = new_name
        self.cmd = cmd


class DeprecatedGroup(click.Group):
    """Group support for deprecating commands."""

    def __init__(self, name=None, commands=None, deprecated=None, **attrs):
        click.Group.__init__(self, name, commands, **attrs)
        self.deprecated = deprecated or []

    def resolve_command(self, ctx, args):
        """Override how the command is resolved to check for deprecated."""
        cmd_name = make_str(args[0])

        for item in self.deprecated:
            if item.old_name == cmd_name:
                message = ('The `{}` command has moved to `{}` '
                           'and will be removed in a future version.')
                logging.warn(message.format(cmd_name, item.new_name))
                return cmd_name, item.cmd, args[1:]
        return click.Group.resolve_command(self, ctx, args)



"""Base translators for translating Grow content."""

import copy
import json
import logging
import os
import threading
import progressbar
import texttable
import yaml
from protorpc import message_types
from protorpc import messages
from protorpc import protojson
from grow.common import progressbar_non
from grow.common import utils
from grow.translators import errors as translator_errors


class TranslatorStat(messages.Message):
    lang = messages.StringField(1)
    num_words = messages.IntegerField(2)
    num_words_translated = messages.IntegerField(3)
    source_lang = messages.StringField(4)
    ident = messages.StringField(5)
    url = messages.StringField(6)
    uploaded = message_types.DateTimeField(7)
    service = messages.StringField(8)
    downloaded = message_types.DateTimeField(9)


def translator_stat_representer(dumper, stat):
    content = json.loads(protojson.encode_message(stat))
    content.pop('lang')  # Exclude from serialization.
    return dumper.represent_mapping('tag:yaml.org,2002:map', content)


yaml.SafeDumper.add_representer(TranslatorStat,
                                translator_stat_representer)


class TranslatorServiceError(Exception):

    def __init__(self, message, ident=None, locale=None):
        if locale:
            new_message = 'Error for locale "{}" -> {}'.format(locale, message)
        elif ident:
            new_message = 'Error for resource "{}" -> {}'.format(
                ident, message)
        else:
            new_message = message
        self.message = new_message
        super(TranslatorServiceError, self).__init__(new_message)


class Translator(object):
    TRANSLATOR_STATS_PATH = '/translators.yaml'
    KIND = None
    has_immutable_translation_resources = False
    has_multiple_langs_in_one_resource = False

    def __init__(self, pod, config=None, project_title=None, instructions=None):
        self.pod = pod
        self.config = config or {}
        self.project_title = project_title or 'Untitled Grow Website'
        self.instructions = instructions

    def _cleanup_locales(self, locales):
        """Certain locales should be ignored."""
        clean_locales = []
        default_locale = self.pod.podspec.default_locale
        skipped = {
            'symlink': set(),
            'po': set(),
        }
        for locale in locales:
            locale_path = os.path.join('translations', str(locale))

            # Silently ignore the default locale.
            if default_locale and str(locale) == str(default_locale):
                continue

            # Ignore the symlinks.
            if os.path.islink(locale_path):
                skipped['symlink'].add(str(locale))
                continue

            # Ignore the locales without a `.PO` file.
            po_path = os.path.join(locale_path, 'LC_MESSAGES', 'messages.po')
            if not self.pod.file_exists(po_path):
                skipped['po'].add(str(locale))
                continue

            clean_locales.append(locale)

        # Summary of skipped files.
        if skipped['symlink']:
            self.pod.logger.info('Skipping: {} (symlinked)'.format(
                ', '.join(sorted(skipped['symlink']))))
        if skipped['po']:
            self.pod.logger.info('Skipping: {} (no `.po` file)'.format(
                ', '.join(sorted(skipped['po']))))

        return clean_locales

    def _download_content(self, stat):
        raise NotImplementedError

    def _log_catalog_changes(self, unchanged_locales, changed_locales):
        if unchanged_locales:
            self.pod.logger.info('No translations updated: {}'.format(
                ', '.join(sorted(unchanged_locales))))
        if changed_locales:
            for locale, value in changed_locales.items():
                self.pod.logger.info('Updated {} of {} translations: {}'.format(
                    value['imported'], value['total'], locale))

    def _upload_catalog(self, catalog, source_lang, prune):
        raise NotImplementedError

    def _upload_catalogs(self, catalogs, source_lang, prune=False):
        raise NotImplementedError

    def _update_acl(self, stat, locale):
        raise NotImplementedError

    def _update_acls(self, stat, locales):
        raise NotImplementedError

    def _update_meta(self, stat, locale, catalog):
        raise NotImplementedError

    def needs_meta_update(self):
        """Allow to be flagged for additional meta update after uploading."""
        return False

    def _get_stats_to_download(self, locales):
        # 'stats' maps the service name to a mapping of languages to stats.
        if not self.pod.file_exists(Translator.TRANSLATOR_STATS_PATH):
            return {}
        stats = self.pod.read_yaml(Translator.TRANSLATOR_STATS_PATH)
        if self.KIND not in stats:
            self.pod.logger.info(
                'Nothing found to download from {}'.format(self.KIND))
            return {}
        stats_to_download = stats[self.KIND]
        if locales:
            stats_to_download = dict([(lang, stat)
                                      for (lang, stat) in stats_to_download.items()
                                      if lang in locales])
        for lang, stat in stats_to_download.items():
            if isinstance(stat, TranslatorStat):
                stat = json.loads(protojson.encode_message(stat))
            stat['lang'] = lang
            stat_message = protojson.decode_message(TranslatorStat, json.dumps(stat))
            stats_to_download[lang] = stat_message
        return stats_to_download

    def download(self, locales, save_stats=True, include_obsolete=False):
        # TODO: Rename to `download_and_import`.
        if not self.pod.file_exists(Translator.TRANSLATOR_STATS_PATH):
            text = 'File {} not found. Nothing to download.'
            self.pod.logger.info(text.format(Translator.TRANSLATOR_STATS_PATH))
            return
        stats_to_download = self._get_stats_to_download(locales)
        if not stats_to_download:
            return
        num_files = len(stats_to_download)
        text = 'Downloading translations: %(value)d/{} (in %(time_elapsed).9s)'
        widgets = [progressbar.FormatLabel(text.format(num_files))]
        bar = progressbar_non.create_progressbar(
            "Downloading translations...", widgets=widgets, max_value=num_files)
        bar.start()
        threads = []
        langs_to_translations = {}
        new_stats = []

        def _do_download(lang, stat):
            try:
                new_stat, content = self._download_content(stat)
            except translator_errors.NotFoundError:
                text = 'No translations to download for: {}'
                self.pod.logger.info(text.format(lang))
                return

            new_stat.uploaded = stat.uploaded  # Preserve uploaded field.
            langs_to_translations[lang] = content
            new_stats.append(new_stat)
        for i, (lang, stat) in enumerate(stats_to_download.items()):
            thread = utils.ProgressBarThread(
                bar, True, target=_do_download, args=(lang, stat))
            threads.append(thread)
            thread.start()
            # Perform the first operation synchronously to avoid oauth2 refresh
            # locking issues.
            if i == 0:
                thread.join()
        for i, thread in enumerate(threads):
            if i > 0:
                thread.join()
        bar.finish()

        has_changed_content = False
        unchanged_locales = []
        changed_locales = {}
        for lang, translations in langs_to_translations.items():
            has_changed_content, imported_translations, total_translations = self.pod.catalogs.import_translations(
                    locale=lang, content=translations,
                    include_obsolete=include_obsolete)
            if imported_translations == 0:
                unchanged_locales.append(lang)
            else:
                changed_locales[lang] = {
                    'imported': imported_translations,
                    'total': total_translations,
                }
            if has_changed_content:
                has_changed_content = True

        if save_stats and has_changed_content:
            self.save_stats(new_stats)
        self._log_catalog_changes(unchanged_locales, changed_locales)
        return new_stats

    def update_acl(self, locales=None):
        locales = locales or self.pod.catalogs.list_locales()
        locales = self._cleanup_locales(locales)
        if not locales:
            self.pod.logger.info('No locales to found to update.')
            return
        stats_to_download = self._get_stats_to_download(locales)
        if not stats_to_download:
            self.pod.logger.info('No documents found to update.')
            return
        if self.has_multiple_langs_in_one_resource:
            self._update_acls(stats_to_download, locales)
            stat = list(stats_to_download.values())[0]
            self.pod.logger.info('ACL updated -> {}'.format(stat.ident))
            return
        threads = []
        for i, (locale, stat) in enumerate(stats_to_download.items()):
            thread = threading.Thread(
                target=self._update_acl, args=(stat, locale))
            threads.append(thread)
            thread.start()
            if i == 0:
                thread.join()
            self.pod.logger.info(
                'ACL updated ({}): {}'.format(stat.lang, stat.url))
        for i, thread in enumerate(threads):
            if i > 0:
                thread.join()

    def update_meta(self, locales=None):
        locales = locales or self.pod.catalogs.list_locales()
        locales = self._cleanup_locales(locales)
        if not locales:
            self.pod.logger.info('No locales to found to update.')
            return
        stats_to_download = self._get_stats_to_download(locales)
        if not stats_to_download:
            self.pod.logger.info('No documents found to update.')
            return
        threads = []
        for i, (locale, stat) in enumerate(stats_to_download.items()):
            catalog_for_meta = self.pod.catalogs.get(locale)
            thread = threading.Thread(
                target=self._update_meta, args=(stat, locale, catalog_for_meta))
            threads.append(thread)
            thread.start()
            if i == 0:
                thread.join()
            self.pod.logger.info('Meta information updated ({}): {}'.format(
                stat.lang, stat.url))
        for i, thread in enumerate(threads):
            if i > 0:
                thread.join()

    def upload(self, locales=None, force=True, verbose=False, save_stats=True,
               prune=False):
        source_lang = self.pod.podspec.default_locale
        locales = locales or self.pod.catalogs.list_locales()
        locales = self._cleanup_locales(locales)
        stats = []
        num_files = len(locales)
        if not locales:
            self.pod.logger.info('No locales to upload.')
            return
        if not force:
            if (self.has_immutable_translation_resources
                    and self.pod.file_exists(Translator.TRANSLATOR_STATS_PATH)):
                text = 'Found existing translator data in: {}'
                self.pod.logger.info(text.format(
                    Translator.TRANSLATOR_STATS_PATH))
                text = 'This will be updated with new data after the upload is complete.'
                self.pod.logger.info(text)
            text = 'Proceed to upload {} translation catalogs?'
            text = text.format(num_files)
            if not utils.interactive_confirm(text):
                self.pod.logger.info('Aborted.')
                return
        if self.has_multiple_langs_in_one_resource:
            catalogs_to_upload = []
            for locale in locales:
                catalog_to_upload = self.pod.catalogs.get(locale)
                if catalog_to_upload:
                    catalogs_to_upload.append(catalog_to_upload)
            stats = self._upload_catalogs(catalogs_to_upload, source_lang,
                                          prune=prune)
        else:
            text = 'Uploading translations: %(value)d/{} (in %(time_elapsed).9s)'
            widgets = [progressbar.FormatLabel(text.format(num_files))]
            bar = progressbar_non.create_progressbar(
                "Uploading translations...", widgets=widgets, max_value=num_files)
            bar.start()
            threads = []

            def _do_upload(locale):
                catalog = self.pod.catalogs.get(locale)
                stat = self._upload_catalog(catalog, source_lang, prune=prune)
                stats.append(stat)
            for i, locale in enumerate(locales):
                thread = utils.ProgressBarThread(
                    bar, True, target=_do_upload, args=(locale,))
                threads.append(thread)
                thread.start()
                # Perform the first operation synchronously to avoid oauth2 refresh
                # locking issues.
                if i == 0:
                    thread.join()
            for i, thread in enumerate(threads):
                if i > 0:
                    thread.join()
            bar.finish()
        stats = sorted(stats, key=lambda stat: stat.lang)
        if verbose:
            self.pretty_print_stats(stats)
        if save_stats:
            self.save_stats(stats)
        return stats

    def save_stats(self, stats):
        """Merges a list of stats into the translator stats file."""
        if self.pod.file_exists(Translator.TRANSLATOR_STATS_PATH):
            content = self.pod.read_yaml(Translator.TRANSLATOR_STATS_PATH)
            create = False
        else:
            content = {}
            create = True
        if self.KIND not in content:
            content[self.KIND] = {}
        for stat in copy.deepcopy(stats):
            content[self.KIND][stat.lang] = stat
        yaml_content = yaml.safe_dump(content, default_flow_style=False)
        self.pod.write_file(Translator.TRANSLATOR_STATS_PATH, yaml_content)
        if create:
            self.pod.logger.info('Saved: {}'.format(
                Translator.TRANSLATOR_STATS_PATH))
        else:
            self.pod.logger.info('Updated: {}'.format(
                Translator.TRANSLATOR_STATS_PATH))

    @classmethod
    def pretty_print_stats(cls, stats):
        table = texttable.Texttable(max_width=0)
        table.set_deco(texttable.Texttable.HEADER)
        rows = []
        rows.append(['Language', 'URL', 'Wordcount'])
        for stat in stats:
            rows.append([stat.lang, stat.url, stat.num_words or '--'])
        table.add_rows(rows)
        logging.info('\n' + table.draw() + '\n')

    def get_edit_url(self, doc):
        if not doc.locale:
            return
        stats = self._get_stats_to_download([doc.locale])
        if doc.locale not in stats:
            return
        stat = stats[doc.locale]
        return stat.url



"""Subcommand for staging pod to remote server."""

import os
import click
from grow.commands import shared
from grow.common import bulk_errors
from grow.common import rc_config
from grow.common import utils
from grow.deployments import stats
from grow.deployments.destinations import base
from grow.deployments.destinations import webreview_destination
from grow.extensions import hooks
from grow.performance import docs_loader
from grow.pods import pods
from grow.rendering import renderer
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.stage')


# pylint: disable=too-many-locals
@click.command()
@shared.pod_path_argument
@click.option('--remote',
              help='WebReview remote address (example: '
                   ' example.com/owner/project). A remote must be specified'
                   ' either using --remote or by configuring a deployment'
                   ' named "webreview" in podspec.yaml.')
@click.option('--file', '--pod-path', 'pod_paths',
              help='Build only pages affected by content files.', multiple=True)
@click.option('--subdomain', help='Assign a subdomain to this build.')
@click.option('--api-key',
              help='API key for authorizing staging to WebReview projects.')
@shared.locale_option(help_text='Filter build routes to specific locale.')
@shared.force_untranslated_option(CFG)
@shared.preprocess_option(CFG)
@shared.threaded_option(CFG)
@shared.work_dir_option
@shared.routes_file_option()
@click.pass_context
def stage(context, pod_path, remote, pod_paths, locale, preprocess, subdomain, api_key,
          force_untranslated, threaded, work_dir, routes_file):
    """Stages a build on a WebReview server."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    auth = context.parent.params.get('auth')
    try:
        pod = pods.Pod(root, storage=storage.FileStorage)
        with pod.profile.timer('grow_stage'):
            deployment = _get_deployment(pod, remote, subdomain, api_key)
            # use the deployment's environment for preprocessing and later
            # steps.
            pod.set_env(deployment.get_env())
            # Always clear the cache when building.
            pod.podcache.reset()
            require_translations = \
                pod.podspec.localization \
                and pod.podspec.localization.get('require_translations', False)
            require_translations = require_translations \
                and not force_untranslated
            if auth:
                deployment.login(auth)
            if preprocess:
                pod.preprocess()
            repo = utils.get_git_repo(pod.root)
            pod.router.use_simple()
            is_partial = bool(pod_paths) or bool(locale)
            if pod_paths:
                pod_paths = [pod.clean_pod_path(path) for path in pod_paths]
                pod.router.add_pod_paths(pod_paths)
            elif routes_file:
                pod.router.from_data(pod.read_json(routes_file))
            else:
                pod.router.add_all()

            if locale:
                pod.router.filter('whitelist', locales=list(locale))

            if not work_dir:
                # Preload the documents used by the paths after filtering.
                docs_loader.DocsLoader.load_from_routes(pod, pod.router.routes)

            paths = pod.router.routes.paths
            stats_obj = stats.Stats(pod, paths=paths)
            content_generator = deployment.dump(
                pod, source_dir=work_dir, use_threading=threaded)
            content_generator = hooks.generator_wrapper(
                pod, 'pre_deploy', content_generator, 'stage')
            deployment.deploy(content_generator, stats=stats_obj, repo=repo,
                              confirm=False, test=False, require_translations=require_translations,
                              is_partial=is_partial)
            pod.podcache.write()
    except bulk_errors.BulkErrors as err:
        # Write the podcache files even when there are rendering errors.
        pod.podcache.write()
        bulk_errors.display_bulk_errors(err)
        raise click.Abort()
    # except base.Error as err:
    #     raise click.ClickException(str(err))
    except pods.Error as err:
        raise click.ClickException(str(err))
    return pod


def _get_deployment(pod, remote, subdomain, api_key):
    if remote:
        dest_class = webreview_destination.WebReviewDestination
        return dest_class(dest_class.Config(remote=remote, name=subdomain))
    else:
        try:
            deployment = pod.get_deployment('webreview')
            if subdomain:
                deployment.config.subdomain = subdomain
            if api_key:
                deployment.webreview.api_key = api_key
            return deployment
        except ValueError:
            text = ('Must provide --remote or specify a deployment named '
                    '"webreview" in podspec.yaml.')
            raise click.ClickException(text)



"""Subcommand to install grow project dependencies."""

import os
import click
from grow.commands import shared
from grow.common import base_config
from grow.common import rc_config
from grow.pods import pods
from grow import storage
from grow.sdk import installer
from grow.sdk import installers


CFG = rc_config.RC_CONFIG.prefixed('grow.install')


@click.command()
@shared.pod_path_argument
@click.option('--gerrit/--no-gerrit', default=None,
              help='Whether to install the Gerrit Code Review commit hook. '
                   'If omitted, Grow will attempt to detect whether there is a '
                   'known Gerrit host amongst the remotes in your repository.')
@click.option('--force', '-f', default=False, is_flag=True,
             help='Whether to force install even when no changes found.')
def install(pod_path, gerrit, force):
    """Checks whether the pod depends on npm, bower, and gulp and installs them
    if necessary. Then, runs install commands. Also optionally installs the
    Gerrit Code Review commit hook."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(
        root, storage=storage.FileStorage, load_extensions=False)
    with pod.profile.timer('grow_install'):
        config = base_config.BaseConfig()
        config.set('gerrit', gerrit)
        config.set('force', force)
        built_in_installers = []
        for installer_class in installers.BUILT_IN_INSTALLERS:
            built_in_installers.append(installer_class(pod, config))
        grow_installer = installer.Installer(built_in_installers, pod)
        grow_installer.run_installers()
    return pod



"""Core extensions."""

from . import deployment_extension
from . import node_extension
from . import pod_extension
from . import podcache_extension
from . import routes_extension

EXTENSIONS = (
    deployment_extension.DeploymentExtension,
    node_extension.NodeExtension,
    pod_extension.PodExtension,
    podcache_extension.PodcacheExtension,
    routes_extension.RoutesExtension,
)



"""Rendering pool for grow documents."""

import random
import threading
from grow.templates import filters
from grow.templates import jinja_dependency
from grow.templates import tags
from grow.templates import tests as jinja_tests


class Error(Exception):
    """Base rendering pool error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class RenderPool(object):
    """Rendering pool for rendering pods."""

    def __init__(self, pod, pool_size=1):
        self.pod = pod
        self._locales = self.pod.list_locales()
        self._pool_size = pool_size
        self._pool = {}
        self._adjust_pool_size(pool_size)

    @property
    def pool_size(self):
        """How many render environments are being used."""
        return self._pool_size

    @pool_size.setter
    def pool_size(self, value):
        """Set how many render environments are being used."""
        if value != self._pool_size:
            self._adjust_pool_size(value)
        self._pool_size = value

    def _adjust_pool_size(self, pool_size):
        """Adjust the pool size.

        Creates a new Jinja Environment for each locale and up to the pool size.

        Also creates a corresponding thread lock that should be used when
        rendering using the environment to prevent the same environment from
        being used in multiple threads at the same time.
        """
        for locale in self._locales:
            locale_str = str(locale)
            if locale_str not in self._pool:
                self._pool[locale_str] = []
            existing_size = len(self._pool[locale_str])
            if pool_size > existing_size:
                for _ in range(pool_size - existing_size):
                    self._pool[locale_str].append({
                        'lock': threading.RLock(),
                        'env': self._create_jinja_env(locale=locale_str),
                    })
            elif pool_size < existing_size:
                self._pool[locale_str] = self._pool[locale_str][:pool_size]

    def _create_jinja_env(self, locale='', root=None):
        kwargs = {
            'autoescape': True,
            'extensions': [
                'jinja2.ext.autoescape',
                'jinja2.ext.do',
                'jinja2.ext.i18n',
                'jinja2.ext.loopcontrols',
                'jinja2.ext.with_',
            ],
            'loader': self.pod.storage.JinjaLoader(self.pod.root if root is None else root),
            'lstrip_blocks': True,
            'trim_blocks': True,
        }
        if self.pod.env.cached:
            kwargs['bytecode_cache'] = self.pod.jinja_bytecode_cache
        kwargs['bytecode_cache'] = self.pod.jinja_bytecode_cache
        kwargs['extensions'].extend(self.pod.list_jinja_extensions())
        env = jinja_dependency.DepEnvironment(**kwargs)
        env.filters.update(filters.create_builtin_filters(env, self.pod, locale=locale))
        env.globals.update(**tags.create_builtin_globals(env, self.pod, locale=locale))
        env.tests.update(jinja_tests.create_builtin_tests())
        self.pod.extensions_controller.trigger('jinja_env_init', env)
        return env

    def custom_jinja_env(self, locale='', root=None):
        """Create a custom jinja env that is not part of the pool."""
        return {
            'lock': threading.RLock(),
            'env': self._create_jinja_env(locale, root),
        }

    def get_jinja_env(self, locale=''):
        """Retrieves a jinja environment and lock to use to render.

        When rendering should use the lock to prevent threading issues.

            with jinja_env['lock']:
                template = jinja_env['env'].get_template('...')
                template.render({})
        """
        locale = str(locale)
        if locale not in self._pool:
            # Add the new locale to the pool.
            self._locales.append(locale)
            self._adjust_pool_size(self._pool_size)

        pool = self._pool[locale]
        # Randomly select which pool item to use. This should distribute
        # rendering fairly evenly across the pool. Since render times are fairly
        # even for most sites this should not be an issue.
        return pool[random.randrange(len(pool))]



"""File watchers for grow."""

import os
from grow.common import colors
from watchdog import events
from watchdog import observers
from . import translation


IGNORED_PATTERNS = [
    '*{sep}extensions{sep}*'.format(sep=os.sep),
    '*{sep}node_modules{sep}*'.format(sep=os.sep),
    '*{sep}build{sep}*'.format(sep=os.sep),
]


class PodFileEventHandler(events.PatternMatchingEventHandler):
    patterns = ['*']
    ignore_directories = True

    def __init__(self, pod, managed_observer, *args, **kwargs):
        self.pod = pod
        self.managed_observer = managed_observer

        # We don't want to write cache files when just serving.
        self.write_cache_file = True
        if 'write_cache_file' in kwargs:
            self.write_cache_file = kwargs.pop('write_cache_file')

        if 'ignore_patterns' not in kwargs:
            kwargs['ignore_patterns'] = IGNORED_PATTERNS
        super(PodFileEventHandler, self).__init__(*args, **kwargs)

    def trigger_file_changed(self, pod_path):
        try:
            self.pod.extensions_controller.trigger(
                'dev_file_change', pod_path, write_cache_file=self.write_cache_file)
        except Exception:  # pylint: disable=broad-except
            # Avoid an inconsistent state where preprocessor doesn't run again
            # if it encounters an exception. https://github.com/grow/grow/issues/528
            colored_pod_path = colors.stylize(pod_path, colors.ERROR)
            self.pod.logger.exception(
                'Found an error -> {}'.format(colored_pod_path))

    def handle(self, event=None):
        if hasattr(event, 'src_path'):
            self.trigger_file_changed(event.src_path[len(self.pod.root):])
        if hasattr(event, 'dest_path'):
            self.trigger_file_changed(event.dest_path[len(self.pod.root):])

        self.managed_observer.reschedule_children()

    def on_any_event(self, event):
        self.handle(event)


class PreprocessorEventHandler(events.PatternMatchingEventHandler):
    num_runs = 0

    def __init__(self, preprocessor, *args, **kwargs):
        self.preprocessor = preprocessor
        if 'ignore_patterns' not in kwargs:
            kwargs['ignore_patterns'] = IGNORED_PATTERNS
        super(PreprocessorEventHandler, self).__init__(*args, **kwargs)

    def handle(self, event=None):
        if event is not None and event.is_directory:
            return
        try:
            if self.num_runs == 0:
                self.preprocessor.first_run()
            else:
                self.preprocessor.run()
        except Exception:  # pylint: disable=broad-except
            # Avoid an inconsistent state where preprocessor doesn't run again
            # if it encounters an exception. https://github.com/grow/grow/issues/81
            text = colors.stylize('Preprocessor error.', colors.ERROR)
            self.preprocessor.pod.logger.exception(text)
        self.num_runs += 1

    def on_created(self, event):
        self.handle(event)

    def on_modified(self, event):
        self.handle(event)


class ManagedObserver(observers.Observer):
    """Observer for pod files."""

    def __init__(self, pod):
        self.pod = pod
        self._preprocessor_watches = []
        self._child_observers = []
        super(ManagedObserver, self).__init__()

    def schedule_podfile(self, write_cache_file=True):
        podfile_handler = PodFileEventHandler(
            self.pod, managed_observer=self, write_cache_file=write_cache_file)
        self.schedule(podfile_handler, path=self.pod.root, recursive=True)

    def schedule_builtins(self):
        preprocessor = translation.TranslationPreprocessor(pod=self.pod)
        self._schedule_preprocessor('/translations/', preprocessor, patterns=['*.po'])

    def schedule_preprocessors(self):
        self._preprocessor_watches = []
        for preprocessor in self.pod.list_preprocessors():
            for path in preprocessor.list_watched_dirs():
                watch = self._schedule_preprocessor(path, preprocessor)
                if watch:
                    self._preprocessor_watches.append(watch)

    def _schedule_preprocessor(self, path, preprocessor, **kwargs):
        try:
            if 'ignore_directories' in kwargs:
                kwargs['ignore_directories'] = [self.pod.abs_path(p)
                                                for p in kwargs['ignore_directories']]
            path = self.pod.abs_path(path)
            if not os.path.isdir(path):
                return None
            handler = PreprocessorEventHandler(preprocessor, **kwargs)
            return self.schedule(handler, path=path, recursive=True)
        except OSError:
            # No directory found.
            return None

    def reschedule_children(self):
        for observer in self._child_observers:
            for watch in observer._preprocessor_watches:
                observer.unschedule(watch)
            observer.schedule_preprocessors()

    def add_child(self, observer):
        self._child_observers.append(observer)
        return observer

    def start(self):
        for observer in self._child_observers:
            observer.start()
        super(ManagedObserver, self).start()

    def stop(self):
        for observer in self._child_observers:
            observer.stop()
        super(ManagedObserver, self).stop()

    def join(self):
        for observer in self._child_observers:
            observer.join()
        super(ManagedObserver, self).join()

    def run_handlers(self):
        for handlers in self._handlers.values():
            for handler in handlers:
                handler.handle()


def create_dev_server_observers(pod, write_cache_file=True):
    """Create the file watchers for the pod."""
    main_observer = ManagedObserver(pod)
    main_observer.schedule_builtins()
    main_observer.schedule_preprocessors()

    podspec_observer = ManagedObserver(pod)
    podspec_observer.schedule_podfile(write_cache_file=write_cache_file)
    podspec_observer.add_child(main_observer)
    podspec_observer.start()
    return main_observer, podspec_observer



"""
Utility for converting Grow sites that use multiple locales in one repository
over to a format that uses separate files for each locale.

This is required for all versions of Grow after 0.1.0.

Usage:

    grow convert --type=content_locale_split
"""

from boltons import iterutils
from collections import OrderedDict
import collections
import copy
import logging
import os
import re
import yaml

try:
    from yaml import CLoader as yaml_Loader
except ImportError:
    from yaml import Loader as yaml_Loader


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class LocaleExistsError(Error):
    pass


class LocaleMissingError(Error):
    pass


BOUNDARY_REGEX = re.compile(r'^-{3,}$', re.MULTILINE)
DEFAULT_LOCALE_REGEX = re.compile(r'^[ ]{2,4}default_locale:[ ]+(.*)')
LOCALE_REGEX = re.compile(r'^\$locale:(.*)')
LOCALES_REGEX = re.compile(r'^\$locales:$')
LOCALIZED_KEY_REGEX = re.compile('(.*)@([^@]+)$')
LOCALIZATION_REGEX = re.compile(r'^\$localization:$')
ARRAY_ITEM_REGEX = re.compile(r'^[ ]*-[ ]+(.*)')
SUB_ITEM_REGEX = re.compile(r'^[ ]{2,4}')
COMBINED_TEMPLATE = '---\n{}\n---\n{}\n'
SINGLE_TEMPLATE = '{}\n'


def _update_deep(orig_dict, new_dict):
    for k, v in new_dict.items():
        if (k in orig_dict and isinstance(orig_dict[k], dict)
                and isinstance(new_dict[k], collections.Mapping)):
            _update_deep(orig_dict[k], new_dict[k])
        else:
            orig_dict[k] = new_dict[k]


class PlainText(object):

    def __init__(self, tag, value):
        self.tag = tag
        self.value = value


class PlainTextYamlLoader(yaml_Loader):

    def construct_plaintext(self, node):
        return PlainText(node.tag, node.value)


class PlainTextYamlDumper(yaml.Dumper):
    pass


def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


def dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def plain_text_representer(dumper, data):
    return dumper.represent_scalar(data.tag, data.value)


# Don't want to actually process the constructors, just keep the values
_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
PlainTextYamlDumper.add_representer(OrderedDict, dict_representer)
PlainTextYamlDumper.add_representer(PlainText, plain_text_representer)
PlainTextYamlLoader.add_constructor(_mapping_tag, dict_constructor)
PlainTextYamlLoader.add_constructor(
    '!_', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.csv', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.doc', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.json', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.static', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.url', PlainTextYamlLoader.construct_plaintext)
PlainTextYamlLoader.add_constructor(
    '!g.yaml', PlainTextYamlLoader.construct_plaintext)


class ConversionDocument(object):

    def __init__(self, pod, file_name, default_locale):
        self.default_locale = default_locale
        self.pod = pod
        self.file_name = file_name
        self.raw_content = pod.read_file(file_name)
        self.normalize_raw_content()

    @staticmethod
    def determine_default_locale(front_matter):
        parsed = yaml.load(front_matter, Loader=PlainTextYamlLoader)
        if '$localization' in parsed:
            return parsed['$localization'].get('default_locale', None)
        return None

    @staticmethod
    def determine_locales(front_matter, default_locale=None,
                          remove_default_locale=True, remove_locales=True):
        if not front_matter:
            return [], None

        parsed = yaml.load(front_matter, Loader=PlainTextYamlLoader)
        if isinstance(parsed, str):
            parsed = OrderedDict()
        locales = parsed.get('$locales', [])

        if '$locale' in parsed:
            locales.append(parsed['$locale'])

        if remove_default_locale:
            if default_locale in locales:
                locales.pop(locales.index(default_locale))
            if '$locales' in parsed and default_locale in parsed['$locales']:
                parsed['$locales'].pop(
                    parsed['$locales'].index(default_locale))
            if '$locale' in parsed and parsed['$locale'] == default_locale:
                del parsed['$locale']

        if remove_locales:
            if '$locales' in parsed:
                del parsed['$locales']
            if '$locale' in parsed:
                del parsed['$locale']

        return locales, yaml.dump(
            parsed, Dumper=PlainTextYamlDumper,
            allow_unicode=True, default_flow_style=False).strip() if parsed else ''

    @staticmethod
    def convert_for_locale(front_matter, locale, base=None):
        if not front_matter:
            parsed = {}
        else:
            parsed = yaml.load(front_matter, Loader=PlainTextYamlLoader)

            def visit(path, key, value):
                if not isinstance(key, str):
                    return key, value
                if key.endswith('@#'):
                    return key, value

                match = LOCALIZED_KEY_REGEX.match(key)
                if not match:
                    return key, value

                base_key = match.group(1)
                locale_from_key = match.group(2)
                if locale_from_key == locale:
                    # If there is a key without the trailing @ then override it.
                    parent = parsed
                    for path_key in path:
                        parent = parent[path_key]
                    if base_key in parent:
                        return base_key, value
                    return '{}@'.format(base_key), value
                return False

            parsed = iterutils.remap(parsed, visit=visit)

        # If there are pre-existing fields, use them as a base for the locale
        # specific values.
        result = base or {}
        _update_deep(result, parsed)

        return result

    @staticmethod
    def format_file(front_matter=None, content=None):
        if front_matter is None or front_matter.strip() == '':
            return SINGLE_TEMPLATE.format(content.lstrip())
        if content is None or content.strip() == '':
            return SINGLE_TEMPLATE.format(front_matter.lstrip())
        return COMBINED_TEMPLATE.format(front_matter.lstrip(), content.lstrip())

    @staticmethod
    def gather_for_locale(front_matter, locale):
        if not front_matter:
            return ''

        parsed = yaml.load(front_matter, Loader=PlainTextYamlLoader)
        locale_extra = OrderedDict()

        def visit(path, key, value):
            if not isinstance(key, str):
                return key, value
            if key.endswith('@#'):
                return key, value

            match = LOCALIZED_KEY_REGEX.match(key)
            if not match:
                return key, value

            base_key = match.group(1)
            locale_from_key = match.group(2)
            if locale_from_key == locale:
                # If there is a key without the trailing @ then override it.
                parent = parsed
                locale_parent = locale_extra

                for path_key in path:
                    parent = parent[path_key]

                    if isinstance(locale_parent, list):
                        locale_parent = locale_parent[path_key]
                    elif path_key not in locale_parent:
                        if isinstance(parent, list):
                            locale_parent[path_key] = copy.deepcopy(parent)
                        else:
                            locale_parent[path_key] = OrderedDict()
                        locale_parent = locale_parent[path_key]
                    else:
                        locale_parent = locale_parent[path_key]

                if base_key in parent:
                    locale_parent[base_key] = value
                else:
                    locale_parent['{}@'.format(base_key)] = value

                if key in locale_parent:
                    locale_parent.pop(key, None)

                return False
            return key, value

        parsed = iterutils.remap(parsed, visit=visit)

        return (yaml.dump(
            parsed, Dumper=PlainTextYamlDumper,
            allow_unicode=True, default_flow_style=False).strip() if parsed else '',
            locale_extra)

    def convert(self):
        # Files with @ in them should already be converted.
        if '@' in self.file_name:
            logging.info(
                'Filename contains a @, skipping: {}'.format(self.file_name))
            return

        # Ignore hidden files.
        if self.file_name.startswith('.'):
            logging.info(
                'Filename starts with ., skipping: {}'.format(self.file_name))
            return

        # Ignore files that don't have an extention
        _, file_extension = os.path.splitext(self.file_name)
        if not file_extension:
            logging.info(
                'Filename does not have an extension, skipping: {}'.format(self.file_name))
            return

        pairs = list(self.split())

        if len(pairs) <= 1:
            logging.info(
                'Single locale detected, skipping: {}'.format(self.file_name))
            return

        logging.info('Converting: {}'.format(self.file_name))
        logging.info(' - Number of content pairs: {}'.format(len(pairs)))

        # Determine if there is a file specific default_locale in first pair.
        default_locale = ConversionDocument.determine_default_locale(
            pairs[0][0]) or self.default_locale
        logging.info(' - Using default_locale: {}'.format(default_locale))

        # Base content will be pruned of localized values that belong in files.
        base_front_matter = pairs[0][0]

        for pair in pairs[1:]:
            locales, _ = ConversionDocument.determine_locales(
                pair[0], default_locale, remove_locales=False,
                remove_default_locale=False)

            if not locales:
                raise LocaleMissingError(
                    'A section in {} is missing a locale and would be lost.'.format(self.file_name))

            # Ensure that there are not existing files for the Locales.
            for locale in locales:
                locale_filename = self.file_name_for_locale(locale)

                if self.pod.file_exists(locale_filename):
                    raise LocaleExistsError(
                        '{} locale section (defined in {}) already has a localized file ({}).\nPlease resolve this confilict and re-run the conversion.'.format(
                            locale, self.file_name, locale_filename))

        # Store each locale contents until the end so we can combine multiple
        # sections that may use the same locale.
        locale_to_content = {}
        locale_to_front_matter = {}

        for pair in pairs[1:]:
            locales, front_matter = ConversionDocument.determine_locales(
                pair[0], default_locale, remove_locales=True,
                remove_default_locale=False)

            for locale in locales:
                locale_to_content[locale] = pair[1]

                if locale in locale_to_front_matter:
                    locale_extra = locale_to_front_matter[locale]
                else:
                    base_front_matter, locale_extra = ConversionDocument.gather_for_locale(
                        base_front_matter, locale)

                # Combine the extra front_matter from the base document with
                # the pair specific front_matter.
                locale_front_matter = ConversionDocument.convert_for_locale(
                    front_matter, locale, base=locale_extra)

                # Store the front matter in case another section adds to it.
                locale_to_front_matter[locale] = locale_front_matter

        # Write the final locale files.
        for locale, locale_front_matter in locale_to_front_matter.items():
            content = locale_to_content.get(locale, None)
            locale_filename = self.file_name_for_locale(locale)
            logging.info('Writing: {}'.format(locale_filename))
            locale_front_matter_dump = yaml.dump(
                locale_front_matter, Dumper=PlainTextYamlDumper, allow_unicode=True,
                default_flow_style=False).strip() if locale_front_matter else ''
            output = ConversionDocument.format_file(
                locale_front_matter_dump, content)
            self.pod.write_file(locale_filename, output)

        # Do the base file after specific tagged fields are removed.
        pair = pairs[0]
        content = pair[1]
        _, base_front_matter = ConversionDocument.determine_locales(
            base_front_matter, default_locale, remove_locales=False,
            remove_default_locale=True)
        logging.info('Writing: {}'.format(self.file_name))
        output = ConversionDocument.format_file(base_front_matter, content)
        self.pod.write_file(self.file_name, output)

    def file_name_for_locale(self, locale_identifier):
        if locale_identifier is None:
            return self.file_name
        file_parts = self.file_name.split('.')
        return '{}@{}.{}'.format(
            '.'.join(file_parts[:-1]), locale_identifier, file_parts[-1])

    def normalize_raw_content(self):
        # Clean and rewrite the yaml files that start with an empty section.
        if self.file_name.endswith('.yaml') and self.raw_content.lstrip().startswith('---'):
            logging.info('Normalizing: {}'.format(self.file_name))
            self.raw_content = self.raw_content.lstrip()[3:].lstrip()
            self.pod.write_file(self.file_name, self.raw_content)

    def split(self):
        parts = BOUNDARY_REGEX.split(self.raw_content)

        # Remove the first, empty list item.
        if parts[0].strip() == '':
            parts.pop(0)

        # Yaml files have no 'content'
        if self.file_name.endswith('.yaml'):
            while parts:
                yield parts.pop(0).strip(), None
        else:
            while parts:
                if len(parts) == 1:
                    yield None, parts.pop(0).strip()
                    break

                front_matter = None
                content = ''
                if parts:
                    front_matter = parts.pop(0).strip() or None
                if parts:
                    content = parts.pop(0).strip() or None
                yield front_matter, content


class Converter(object):

    @staticmethod
    def convert(pod):
        default_locale = pod.podspec.default_locale

        logging.info('Using default locale: {}'.format(default_locale))

        # Go through each document and convert to the updated format.
        for root, dirs, files in pod.walk('/content'):
            pod_dir = root.replace(pod.root, '')

            for file_name in files:
                doc = ConversionDocument(
                    pod, os.path.join(pod_dir, file_name), default_locale)
                try:
                    doc.convert()
                except:
                    print('Error trying to convert: {}'.format(
                        os.path.join(pod_dir, file_name)))
                    raise
